# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import time
import hashlib
import pathlib
from unittest.mock import patch, MagicMock, call
import os
import uuid

import pytest
import tomlkit
import tomllib

from dorsal.api import file as file_api
from dorsal.common.exceptions import (
    DorsalError,
    DorsalClientError,
    DorsalConfigError,
    NotFoundError,
    ConflictError,
)
from dorsal.common.model import AnnotationModel, AnnotationModelSource
from dorsal.common.validators import Pagination
from dorsal.client.validators import (
    FileDeleteResponse,
    FileIndexResponse,
    FileTagResponse,
)
from dorsal.file.validators.base import FileCoreValidationModel
from dorsal.file.validators.file_record import (
    FileRecordDateTime,
    FileSearchResponse,
    Annotations,
    Annotation_Base,
    AnnotationStub,
    GenericFileAnnotation,
    NewFileTag,
)


class MockFileRecord:
    def __init__(self, hash_value, name):
        self.hash = hash_value
        self.name = name

    def model_dump(self, **kwargs):
        return {"hash": self.hash, "name": self.name}

    def model_dump_json(self, **kwargs):
        import json

        return json.dumps(self.model_dump())


@pytest.fixture
def mock_shared_client():
    """Mocks the get_shared_dorsal_client function."""
    with patch("dorsal.session.get_shared_dorsal_client") as mock_get:
        mock_client = MagicMock()
        mock_get.return_value = mock_client
        yield mock_client


@pytest.fixture
def mock_cache():
    """Mocks the get_shared_cache function."""
    with patch("dorsal.session.get_shared_cache") as mock_get:
        mock_cache_instance = MagicMock()

        mock_cache_instance.get_hash.return_value = None
        mock_get.return_value = mock_cache_instance
        yield mock_cache_instance


@pytest.fixture
def mock_pagination_json() -> dict:
    """Provides a valid dictionary for a Pagination API response."""
    return {
        "current_page": 1,
        "record_count": 1,
        "page_count": 1,
        "per_page": 50,
        "has_next": False,
        "has_prev": False,
        "start_index": 0,
        "end_index": 0,
        "total_items": 1,
        "total_pages": 1,
    }


@pytest.fixture
def mock_jinja_env():
    """Mocks the Jinja2 environment to avoid needing real templates."""
    with patch("jinja2.Environment") as mock_env:
        mock_template = MagicMock()
        mock_template.render.return_value = "<html>Rendered Report</html>"
        mock_env.return_value.get_template.return_value = mock_template
        yield mock_env


class MockLocalFile:
    """A simple mock to stand in for the real LocalFile object."""

    def __init__(self, path, hash_value):
        self.path = path
        self.hash = hash_value


class MyTestModel(AnnotationModel):
    my_field: str

    @classmethod
    def process(cls, file, **kwargs) -> "MyTestModel":
        return MyTestModel(my_field="test")


@pytest.fixture
def mock_metadata_reader():
    """Mocks the get_metadata_reader function for dependency injection."""
    with patch("dorsal.api.file.get_metadata_reader") as mock_get:
        mock_reader_instance = MagicMock()
        mock_get.return_value = mock_reader_instance
        yield mock_reader_instance


_DUMMY_CLIENT = MagicMock()
_DUMMY_SHA256 = "a" * 64


def test_identify_file_success_sha256(mock_shared_client, tmp_path):
    """Test identifying a file successfully using its SHA-256 hash."""

    file = tmp_path / "test.txt"
    file.write_text("content")

    expected_record = MockFileRecord(
        hash_value="ed7002b439e9ac845f22357d822bac1444730fbdb6016d3ec9432297b9ec9f73",
        name="test.txt",
    )
    mock_shared_client.download_file_record.return_value = expected_record

    result = file_api.identify_file(str(file), quick=False)

    mock_shared_client.download_file_record.assert_called_once()
    assert "SHA-256:" in mock_shared_client.download_file_record.call_args[1]["hash_string"]
    assert result.hash == expected_record.hash


def test_identify_file_quick_hash_fallback(mock_shared_client, tmp_path):
    """Test that identify_file falls back to SHA-256 if a quick hash is not found."""

    file = tmp_path / "large_file.bin"
    file.write_bytes(b"\0" * (32 * 1024 * 1024))

    sha256_record = MockFileRecord(hash_value="...", name="large_file.bin")
    mock_shared_client.download_file_record.side_effect = [
        NotFoundError("Quick hash not found", request_url="http://dorsalhub.test/missing-thing"),
        sha256_record,
    ]

    result = file_api.identify_file(str(file), quick=True)

    assert mock_shared_client.download_file_record.call_count == 2
    assert "QUICK:" in mock_shared_client.download_file_record.call_args_list[0][1]["hash_string"]
    assert "SHA-256:" in mock_shared_client.download_file_record.call_args_list[1][1]["hash_string"]
    assert result == sha256_record


def test_identify_file_not_found_raises_dorsal_error(mock_shared_client, tmp_path):
    """Test that a NotFoundError from the client is wrapped with a helpful message."""
    file = tmp_path / "test.txt"
    file.write_text("content")

    mock_shared_client.download_file_record.side_effect = NotFoundError(
        "Original not found", request_url="http://dorsalhub.test/missing-thing"
    )

    with pytest.raises(DorsalClientError):
        file_api.identify_file(str(file), quick=False)


@patch("dorsal.client.dorsal_client.read_api_key", return_value="dummy-key-for-test")
def test_identify_file_os_error_raises_file_not_found(mock_read_api_key):
    """Test that a non-existent file path raises FileNotFoundError."""
    with pytest.raises(FileNotFoundError):
        file_api.identify_file("path/to/non_existent_file.txt")


def test_get_dorsal_file_record_success(mock_shared_client):
    """Test a successful agnostic search for a file record."""
    hash_str = "some_hash"
    expected_record = MockFileRecord(hash_value=hash_str, name="found_file.zip")
    mock_shared_client.download_file_record.return_value = expected_record

    result = file_api.get_dorsal_file_record(hash_str, mode="pydantic")

    mock_shared_client.download_file_record.assert_called_once_with(hash_string=hash_str, private=None)
    assert result == expected_record


@pytest.mark.parametrize("mode, expected_type", [("dict", dict), ("json", str)])
def test_get_dorsal_file_record_modes(mock_shared_client, mode, expected_type):
    """Test the 'mode' parameter returns the correct data type."""
    hash_str = "some_hash"
    mock_record = MockFileRecord(hash_value=hash_str, name="file.txt")
    mock_shared_client.download_file_record.return_value = mock_record

    result = file_api.get_dorsal_file_record(hash_str, mode=mode)

    assert isinstance(result, expected_type)


@patch("dorsal.client.DorsalClient")
def test_get_dorsal_file_record_with_api_key(mock_client_class, mock_shared_client):
    """Test that providing an api_key creates a temporary client."""

    mock_temp_client = MagicMock()
    mock_client_class.return_value = mock_temp_client

    file_api.get_dorsal_file_record("some_hash", api_key="temp_key_123")

    mock_shared_client.download_file_record.assert_not_called()
    mock_client_class.assert_called_once_with(api_key="temp_key_123")
    mock_temp_client.download_file_record.assert_called_once()


def test_delete_dorsal_file_record_success(mock_shared_client):
    """Test successful agnostic deletion of a file record."""
    file_hash = "a" * 64
    mock_delete_response = FileDeleteResponse(file_deleted=1)
    mock_shared_client.delete_file.return_value = mock_delete_response

    result = file_api._delete_dorsal_file_record(file_hash)

    mock_shared_client.delete_file.assert_called_once_with(
        file_hash=file_hash, record="all", tags="all", annotations="all"
    )
    assert result.file_deleted == 1


def test_delete_dorsal_file_record_invalid_hash():
    """Test that _delete_dorsal_file_record raises ValueError for an invalid hash."""

    with pytest.raises(ValueError, match="file_hash must be a valid SHA-256 hash"):
        file_api._delete_dorsal_file_record("not-a-hash")


@patch("dorsal.file.dorsal_file.LocalFile")
def test_index_file_success(mock_local_file_cls, tmp_path):
    """Test successful indexing of a single file."""
    file = tmp_path / "things_and_stuff.txt"
    file.write_text("content")

    mock_instance = mock_local_file_cls.return_value
    mock_index_response = FileIndexResponse(total=1, success=1, error=0, unauthorized=0, results=[])
    mock_instance.push.return_value = mock_index_response

    result = file_api.index_file(str(file), public=False, use_cache=False)

    assert result == mock_index_response
    mock_local_file_cls.assert_called_with(file_path=str(file), use_cache=False)
    mock_instance.push.assert_called_with(public=False, api_key=None, strict=False)


def test_add_tag_to_file_success(mock_shared_client):
    """Test successfully adding a tag via the high-level function."""
    file_hash = "a" * 64
    tag_name = "status"
    tag_value = "reviewed"
    mock_response = FileTagResponse(success=True, hash=file_hash)
    mock_shared_client.add_tags_to_file.return_value = mock_response

    result = file_api.add_tag_to_file(file_hash, tag_name, tag_value, public=False)

    mock_shared_client.add_tags_to_file.assert_called_once()
    call_args = mock_shared_client.add_tags_to_file.call_args[1]
    assert call_args["file_hash"] == file_hash
    assert isinstance(call_args["tags"][0], NewFileTag)
    assert call_args["tags"][0].name == tag_name
    assert call_args["tags"][0].value == tag_value
    assert call_args["tags"][0].private is True
    assert result.success is True


@patch("dorsal.api.file.add_tag_to_file")
def test_add_label_to_file_delegation(mock_add_tag):
    """Test that add_label_to_file correctly wraps add_tag_to_file."""
    file_hash = "a" * 64
    label = "urgent"
    api_key = "secret_key"

    mock_response = MagicMock()
    mock_add_tag.return_value = mock_response

    result = file_api.add_label_to_file(hash_string=file_hash, label=label, api_key=api_key)

    mock_add_tag.assert_called_once_with(
        hash_string=file_hash, name="label", value=label, public=False, api_key=api_key
    )
    assert result == mock_response


def test_remove_tag_from_file_success(mock_shared_client):
    """Test successfully removing a tag via the high-level function."""
    file_hash = _DUMMY_SHA256
    tag_id = "t_12345"

    mock_shared_client.delete_tag.return_value = None

    result = file_api.remove_tag_from_file(file_hash, tag_id)

    mock_shared_client.delete_tag.assert_called_once_with(file_hash=file_hash, tag_id=tag_id)
    assert result is None


@patch("dorsal.api.file.get_metadata_reader")
def test_search_user_files_success(mock_get_reader, mock_pagination_json):
    """Test a successful search scoped to the user."""
    mock_client = MagicMock()
    mock_get_reader.return_value._client = mock_client

    mock_pagination = mock_pagination_json
    mock_response = FileSearchResponse(api_version="1.0", pagination=mock_pagination, results=[], errors=[])
    mock_client.search_files.return_value = mock_response

    query = "extension:pdf"
    result = file_api.search_user_files(query)

    mock_client.search_files.assert_called_once()
    call_args = mock_client.search_files.call_args[1]
    assert call_args["q"] == query
    assert call_args["scope"] == "user"
    assert result == mock_response


@patch("dorsal.api.file.get_metadata_reader")
def test_search_global_files_success(mock_get_reader, mock_pagination_json):
    """Test a successful search scoped globally."""
    mock_client = MagicMock()
    mock_get_reader.return_value._client = mock_client

    mock_pagination = mock_pagination_json
    mock_response = FileSearchResponse(api_version="1.0", pagination=mock_pagination, results=[], errors=[])
    mock_client.search_files.return_value = mock_response

    query = "tag:research"
    result = file_api.search_global_files(query)

    mock_client.search_files.assert_called_once()
    call_args = mock_client.search_files.call_args[1]
    assert call_args["q"] == query
    assert call_args["scope"] == "global"
    assert result == mock_response


def test_scan_file_success(mock_metadata_reader):
    """Test that scan_file correctly calls the underlying reader method."""
    file_path = "/tmp/test.txt"
    mock_local_file = MockLocalFile(path=file_path, hash_value="a" * 64)
    mock_metadata_reader.scan_file.return_value = mock_local_file

    result = file_api.scan_file(file_path, use_cache=False)

    mock_metadata_reader.scan_file.assert_called_once_with(file_path=file_path, skip_cache=True, follow_symlinks=True)
    assert result == mock_local_file


def test_scan_directory_success(mock_metadata_reader):
    """Test that scan_directory correctly calls the underlying reader method."""
    dir_path = "/tmp/docs"
    mock_files = [MockLocalFile(path="/tmp/docs/a.txt", hash_value="a" * 64)]
    mock_metadata_reader.scan_directory.return_value = mock_files

    result = file_api.scan_directory(dir_path, recursive=True, use_cache=True)

    mock_metadata_reader.scan_directory.assert_called_once_with(
        dir_path=dir_path, recursive=True, skip_cache=False, follow_symlinks=True
    )
    assert result == mock_files


@patch("dorsal.api.file.get_metadata_reader")
def test_index_directory_success(mock_get_reader):
    """Test the orchestration logic of the index_directory function."""
    dir_path = "/tmp/assets"

    mock_reader = MagicMock()
    mock_get_reader.return_value = mock_reader

    mock_summary = {"total_records": 1, "success": 1, "failed": 0, "batches": [], "errors": []}
    mock_reader.index_directory.return_value = mock_summary
    summary = file_api.index_directory(dir_path, public=False)

    mock_get_reader.assert_called_once()

    mock_reader.index_directory.assert_called_once_with(
        dir_path=dir_path, recursive=False, public=False, skip_cache=False, fail_fast=True
    )

    assert summary == mock_summary


def test_find_duplicates_success(fs):
    """Test finding duplicate files in a directory.
    'fs' is a fixture provided by the pyfakefs library.
    """

    fs.create_file("/test/unique1.txt", contents="abc")
    fs.create_file("/test/duplicate1.txt", contents="12345")
    fs.create_dir("/test/subdir")
    fs.create_file("/test/subdir/duplicate2.txt", contents="12345")
    fs.create_file("/test/subdir/unique2.txt", contents="xyz")

    correct_hash = hashlib.sha256(b"12345").hexdigest()

    with patch("dorsal.api.file.get_shared_cache") as mock_get_cache:
        mock_get_cache.return_value.get_hash.return_value = None
        result = file_api.find_duplicates("/test", recursive=True, mode="sha256")

    assert result["total_sets"] == 1
    duplicate_set = result["duplicate_sets"][0]
    assert duplicate_set["count"] == 2
    assert duplicate_set["hash"] == correct_hash

    expected_paths = {
        os.path.normpath("/test/duplicate1.txt"),
        os.path.normpath("/test/subdir/duplicate2.txt"),
    }
    assert set(duplicate_set["paths"]) == expected_paths


def test_find_duplicates_no_results(fs):
    """Test that find_duplicates returns an empty sets list when no duplicates are found."""
    fs.create_file("/test/unique1.txt", contents="abc")
    fs.create_file("/test/unique2.txt", contents="12345")

    result = file_api.find_duplicates("/test", mode="sha256")

    assert result["total_sets"] == 0
    assert result["duplicate_sets"] == []
    assert result["path"] == os.path.normpath("/test")


def test_get_directory_info_success(fs):
    """Test getting a statistical summary of a directory."""
    fs.create_file("/test/file1.txt", contents="a" * 100)
    fs.create_file("/test/file2.bin", contents="b" * 200)
    fs.create_dir("/test/subdir")
    fs.create_file("/test/subdir/file3.txt", contents="c" * 50)

    result = file_api.get_directory_info("/test", recursive=True)

    overall = result["overall"]
    assert overall["total_files"] == 3
    assert overall["total_dirs"] == 1
    assert overall["total_size"] == 350

    assert overall["largest_file"]["path"] == os.path.normpath("/test/file2.bin")
    assert overall["smallest_file"]["path"] == os.path.normpath("/test/subdir/file3.txt")
    assert len(result["by_type"]) > 0


@patch("dorsal.api.file.resolve_template_path")
def test_generate_html_file_report_success(mock_resolve, mock_jinja_env, mock_metadata_reader, tmp_path):
    """Test generating a file report with mocked Jinja2."""

    file_path = tmp_path / "report_target.txt"
    output_path = tmp_path / "report.html"
    file_path.write_text("data")

    mock_resolve.return_value = (pathlib.Path("default.html"), "/templates/base")

    mock_local_file = MagicMock()
    mock_local_file._file_path = str(file_path)
    mock_local_file.date_created = datetime.datetime.now()
    mock_local_file.date_modified = datetime.datetime.now()

    mock_local_file.to_dict.return_value = {
        "annotations": {"file/base": {"record": {"name": "report_target.txt", "size": 4}}}
    }

    html_out = file_api.generate_html_file_report(
        str(file_path), local_file=mock_local_file, output_path=str(output_path)
    )

    assert html_out is None
    assert output_path.exists()
    assert output_path.read_text(encoding="utf-8") == "<html>Rendered Report</html>"


@patch("dorsal.api.file.resolve_template_path")
@patch("dorsal.common.config.get_collection_report_panel_config")
def test_generate_html_directory_report_success(mock_panel_config, mock_resolve, mock_jinja_env, tmp_path):
    """Test generating a directory dashboard."""
    dir_path = tmp_path / "assets"
    dir_path.mkdir()
    output_path = tmp_path / "dashboard.html"

    mock_panel_config.return_value = {"overview": True, "duplicates": False}

    mock_resolve.return_value = (pathlib.Path("dashboard.html"), "/templates/base")

    mock_collection = MagicMock()
    mock_collection.to_dict.return_value = {"files": []}

    with patch.dict("dorsal.file.utils.reports.REPORT_DATA_GENERATORS", {"overview": lambda c: "overview_data"}):
        html_out = file_api.generate_html_directory_report(
            str(dir_path), local_collection=mock_collection, output_path=str(output_path)
        )

    assert html_out is None
    assert output_path.exists()

    call_args = mock_jinja_env.return_value.get_template.return_value.render.call_args
    context = call_args[0][0]
    assert context["report_title"] == "Directory Report: assets"
    assert context["panels"][0]["id"] == "overview"


def test_get_directory_info_detailed_metrics(fs):
    """
    Tests the detailed metric collection: permissions, dates, and media types.
    Using pyfakefs to simulate file stats accurately.
    """
    fs.create_file("/data/ro.txt", contents="read only", st_mode=0o444)
    fs.create_file("/data/exe.sh", contents="exec", st_mode=0o755)
    fs.create_file("/data/old.txt", contents="old")
    fs.create_file("/data/new.txt", contents="new")

    now = time.time()

    os.utime("/data/ro.txt", (now - 500, now - 500))
    os.utime("/data/exe.sh", (now - 500, now - 500))
    os.utime("/data/old.txt", (now - 1000, now - 1000))
    os.utime("/data/new.txt", (now, now))

    with patch("dorsal.api.file.get_media_type", side_effect=lambda p, e: "text/plain"):
        result = file_api.get_directory_info("/data", recursive=False, media_type=True)

    overall = result["overall"]

    assert overall["permissions"]["executable"] >= 1
    assert overall["permissions"]["read_only"] >= 1

    assert overall["newest_mod_file"]["path"] == os.path.normpath("/data/new.txt")
    assert overall["oldest_mod_file"]["path"] == os.path.normpath("/data/old.txt")


def test_get_directory_info_progress_integration(fs):
    """Tests that get_directory_info interacts correctly with a Console progress bar."""
    fs.create_file("/data/f1.txt")
    fs.create_file("/data/f2.txt")

    mock_console = MagicMock()

    with patch("dorsal.api.file._create_rich_progress") as mock_create_progress:
        mock_progress_instance = MagicMock()
        mock_create_progress.return_value = mock_progress_instance

        file_api.get_directory_info("/data", progress_console=mock_console)

        mock_create_progress.assert_called_once()
        mock_progress_instance.add_task.assert_called()

        assert mock_progress_instance.update.call_count >= 2


def test_find_duplicates_quick_internal_logic(fs):
    """
    Target lines 1786-1856 (_find_duplicates_quick).
    We setup 3 files to hit 3 specific code paths in the loop:
    1. 'cached.txt' -> Hits the Cache (skips calculation).
    2. 'quick.txt'  -> Misses Cache, Hits Quick Hash.
    3. 'fallback.txt' -> Misses Cache, Misses Quick Hash, Falls back to SHA256.
    """
    fs.create_file("/cached.txt", contents="A")
    fs.create_file("/quick.txt", contents="B")
    fs.create_file("/fallback.txt", contents="C")

    mock_cache = MagicMock()

    def mock_get_hash(path, hash_function):
        if "cached.txt" in path and hash_function == "QUICK":
            return "cached_hash_val"
        return None

    mock_cache.get_hash.side_effect = mock_get_hash

    with (
        patch("dorsal.api.file.get_shared_cache", return_value=mock_cache),
        patch("dorsal.api.file.get_quick_hash") as mock_quick,
        patch("dorsal.api.file.get_sha256_hash") as mock_sha,
    ):

        def quick_side_effect(path, **kwargs):
            if "quick.txt" in str(path):
                return "quick_hash_val"
            return None

        mock_quick.side_effect = quick_side_effect
        mock_sha.return_value = "sha_fallback_val"

        result = file_api.find_duplicates("/", mode="quick", use_cache=True)

    assert mock_cache.get_hash.called

    assert mock_quick.call_count == 2

    assert mock_sha.call_count == 1

    assert result["hashes_from_cache"] == 1


@patch("dorsal.api.file.resolve_template_path")
@patch("dorsal.common.config.get_collection_report_panel_config")
def test_generate_html_directory_report_panels(mock_panel_config, mock_resolve, mock_metadata_reader, tmp_path):
    """
    Target lines 2337-2414 (generate_html_directory_report).
    We define a config with TWO panels:
    1. 'overview' -> Exists in generators (Hits 'if generator_func')
    2. 'missing_panel' -> Does NOT exist (Hits 'else: logger.warning')
    """
    dir_path = tmp_path / "report_test"
    dir_path.mkdir()
    output_path = tmp_path / "report.html"

    mock_panel_config.return_value = {"overview": True, "missing_panel": True}

    mock_resolve.return_value = (pathlib.Path("default.html"), "/templates/base")

    mock_collection = MagicMock()
    mock_collection.to_dict.return_value = {"files": []}

    fake_generators = {"overview": lambda c: "data"}

    with (
        patch.dict("dorsal.file.utils.reports.REPORT_DATA_GENERATORS", fake_generators),
        patch("jinja2.Environment") as mock_env,
    ):
        mock_template = MagicMock()
        mock_template.render.return_value = "<html></html>"
        mock_env.return_value.get_template.return_value = mock_template

        file_api.generate_html_directory_report(
            str(dir_path), local_collection=mock_collection, output_path=str(output_path)
        )

        call_args = mock_template.render.call_args
        context = call_args[0][0]

        assert len(context["panels"]) == 1
        assert context["panels"][0]["id"] == "overview"


@pytest.fixture
def dummy_file_record_dt_with_stubs() -> FileRecordDateTime:
    """Provides a valid FileRecordDateTime populated with a custom AnnotationStub."""
    base_record = FileCoreValidationModel(
        hash="a" * 64, name="test.txt", extension=".txt", size=100, media_type="text/plain"
    )
    file_base = Annotation_Base(
        record=base_record, source=AnnotationModelSource(type="Model", id="file/base", version="1.0.0")
    )
    annotations = Annotations(file_base=file_base)

    stub = AnnotationStub(
        hash="a" * 64,
        id=uuid.uuid4(),
        source=AnnotationModelSource(type="Model", id="custom_model"),
        user_id=1,
        date_modified=datetime.datetime.now(datetime.UTC),
    )

    annotations.__pydantic_extra__ = {"open/custom": [stub]}

    return FileRecordDateTime(
        hash="a" * 64,
        date_created=datetime.datetime.now(datetime.UTC),
        date_modified=datetime.datetime.now(datetime.UTC),
        annotations=annotations,
    )


def test_get_file_annotation_success(mock_shared_client):
    """Test directly fetching a specific annotation by ID."""

    dummy_response = GenericFileAnnotation(custom_field="hydrated_data")
    mock_shared_client.get_file_annotation.return_value = dummy_response

    result = file_api.get_file_annotation(hash_string="a" * 64, annotation_id="123")

    mock_shared_client.get_file_annotation.assert_called_once_with(file_hash="a" * 64, annotation_id="123")
    assert result == dummy_response


def test_get_file_annotation_not_found(mock_shared_client):
    """Test that a 404 is caught and wrapped with a clear error message."""
    mock_not_found = NotFoundError(message="404", request_url="http://test")

    mock_shared_client.get_file_annotation.side_effect = mock_not_found

    with pytest.raises(DorsalClientError, match="Annotation '123' not found for file"):
        file_api.get_file_annotation(hash_string="a" * 64, annotation_id="123")


def test_get_latest_file_annotation_success(mock_shared_client, dummy_file_record_dt_with_stubs):
    """Test retrieving and automatically hydrating the latest annotation."""

    mock_shared_client.download_file_record.return_value = dummy_file_record_dt_with_stubs

    dummy_hydrated = GenericFileAnnotation(custom_data="full_content")
    mock_shared_client.get_file_annotation.return_value = dummy_hydrated

    result = file_api.get_latest_file_annotation(hash_string="a" * 64, schema_id="open/custom")

    mock_shared_client.download_file_record.assert_called_once()
    mock_shared_client.get_file_annotation.assert_called_once()
    assert result == dummy_hydrated


def test_get_latest_file_annotation_formatting(mock_shared_client, dummy_file_record_dt_with_stubs):
    """Test that the hydrated data respects the 'mode' formatting argument."""
    mock_shared_client.download_file_record.return_value = dummy_file_record_dt_with_stubs

    dummy_hydrated = GenericFileAnnotation(custom_data="full_content")
    mock_shared_client.get_file_annotation.return_value = dummy_hydrated

    result_dict = file_api.get_latest_file_annotation(hash_string="a" * 64, schema_id="open/custom", mode="dict")
    assert isinstance(result_dict, dict)
    assert result_dict["custom_data"] == "full_content"


def test_get_latest_file_annotation_not_found(mock_shared_client, dummy_file_record_dt_with_stubs):
    """Test behavior when the requested schema doesn't exist on the file."""
    mock_shared_client.download_file_record.return_value = dummy_file_record_dt_with_stubs

    with pytest.raises(NotFoundError, match="No annotations found for schema 'open/missing'"):
        file_api.get_latest_file_annotation(hash_string="a" * 64, schema_id="open/missing")


def test_get_file_annotations_summary(mock_shared_client, dummy_file_record_dt_with_stubs):
    """Test retrieving a lightweight summary list of stubs."""
    mock_shared_client.download_file_record.return_value = dummy_file_record_dt_with_stubs

    result = file_api.get_file_annotations_summary(hash_string="a" * 64, schema_id="open/custom")

    assert isinstance(result, list)
    assert len(result) == 1

    stub_summary = result[0]
    assert "id" in stub_summary
    assert "source" in stub_summary
    assert "url" in stub_summary

    mock_shared_client.get_file_annotation.assert_not_called()


@patch("dorsal.api.file.get_sha256_hash")
def test_identify_file_quick_hash_collision(mock_sha256, mock_shared_client, tmp_path):
    """Test that a ConflictError on a quick hash falls back to SHA-256."""
    file = tmp_path / "large_file.bin"
    file.write_bytes(b"\0" * (32 * 1024 * 1024))

    expected_record = MockFileRecord(hash_value="sha256-hash-value", name="large_file.bin")

    mock_shared_client.download_file_record.side_effect = [
        ConflictError("Collision detected"),
        expected_record,
    ]
    mock_sha256.return_value = "sha256-hash-value"

    result = file_api.identify_file(str(file), quick=True)

    assert mock_shared_client.download_file_record.call_count == 2
    assert "QUICK:" in mock_shared_client.download_file_record.call_args_list[0][1]["hash_string"]
    assert "SHA-256:" in mock_shared_client.download_file_record.call_args_list[1][1]["hash_string"]
    assert result == expected_record


@patch("dorsal.api.file.get_sha256_hash")
def test_identify_file_sha256_failure(mock_sha256, mock_shared_client, tmp_path):
    """Test behavior when SHA-256 generation fails (returns None)."""
    file = tmp_path / "test.txt"
    file.write_text("content")

    mock_sha256.return_value = None

    with pytest.raises(DorsalError, match="Could not generate SHA-256 hash for file"):
        file_api.identify_file(str(file), quick=False)


def test_identify_file_unexpected_error(mock_shared_client, tmp_path):
    """Test that unexpected exceptions are wrapped in a DorsalError."""
    file = tmp_path / "test.txt"
    file.write_text("content")

    mock_shared_client.download_file_record.side_effect = Exception("Random crash")

    with pytest.raises(DorsalError, match="An unexpected error occurred while identifying file"):
        file_api.identify_file(str(file), quick=False)


@pytest.mark.parametrize("invalid_hash", ["", "   ", None])
def test_get_dorsal_file_record_empty_hash(invalid_hash, mock_shared_client):
    """Test that empty or whitespace-only hashes raise a ValueError."""
    with pytest.raises(ValueError, match="hash_string must be a non-empty string"):
        file_api.get_dorsal_file_record(invalid_hash)  # type: ignore


def test_get_dorsal_file_record_not_found(mock_shared_client):
    """Test that NotFoundError is caught and context is added to the message."""
    mock_not_found = NotFoundError(message="Original 404", request_url="http://test")

    mock_shared_client.download_file_record.side_effect = mock_not_found

    with pytest.raises(DorsalClientError, match="File not found in 'Agnostic.*' scope for hash 'missing_hash'"):
        file_api.get_dorsal_file_record("missing_hash")


def test_get_dorsal_file_record_unexpected_error(mock_shared_client):
    """Test that generic exceptions are wrapped in DorsalError."""
    mock_shared_client.download_file_record.side_effect = Exception("Network timeout")

    with pytest.raises(DorsalError, match="An unexpected error occurred while getting metadata for hash 'crash_hash'"):
        file_api.get_dorsal_file_record("crash_hash")


def test_get_dorsal_file_record_invalid_mode(mock_shared_client):
    """Test that an invalid mode literal raises a ValueError."""
    mock_shared_client.download_file_record.return_value = MockFileRecord("abc", "test.txt")

    with pytest.raises(ValueError, match="Invalid mode: 'xml'"):
        file_api.get_dorsal_file_record("abc", mode="xml")  # type: ignore


def test_get_file_annotation_generic_client_error(mock_shared_client):
    """Test that a non-404 DorsalClientError is re-raised directly."""

    mock_shared_client.get_file_annotation.side_effect = DorsalClientError("Permission Denied")

    with pytest.raises(DorsalClientError, match="Permission Denied"):
        file_api.get_file_annotation("anno_123", "hash_abc")


def test_get_file_annotation_unexpected_error(mock_shared_client):
    """Test that generic exceptions are wrapped in DorsalError."""
    mock_shared_client.get_file_annotation.side_effect = Exception("System fault")

    with pytest.raises(DorsalError, match="Unexpected error fetching annotation 'anno_123'"):
        file_api.get_file_annotation("anno_123", "hash_abc")


def test_get_file_annotation_invalid_mode(mock_shared_client):
    """Test that an invalid mode raises a ValueError."""
    mock_shared_client.get_file_annotation.return_value = GenericFileAnnotation()

    with pytest.raises(ValueError, match="Invalid mode: 'yaml'"):
        file_api.get_file_annotation("anno_123", "hash_abc", mode="yaml")  # type: ignore


@patch("dorsal.file.dorsal_file.DorsalFile")
def test_get_latest_file_annotation_generic_client_error(mock_dorsal_file_cls, mock_shared_client):
    """Test that DorsalClientError propagates up."""
    mock_instance = mock_dorsal_file_cls.return_value
    mock_instance.get_latest_annotation.side_effect = DorsalClientError("Rate Limited")

    with pytest.raises(DorsalClientError, match="Rate Limited"):
        file_api.get_latest_file_annotation("hash_abc", "open/schema")


@patch("dorsal.file.dorsal_file.DorsalFile")
def test_get_latest_file_annotation_unexpected_error(mock_dorsal_file_cls, mock_shared_client):
    """Test that unexpected exceptions are wrapped in a DorsalError."""
    mock_instance = mock_dorsal_file_cls.return_value
    mock_instance.get_latest_annotation.side_effect = TypeError("Bad serialization")

    with pytest.raises(DorsalError, match="Unexpected error fetching latest 'open/schema' annotation:"):
        file_api.get_latest_file_annotation("hash_abc", "open/schema")


@patch("dorsal.file.dorsal_file.DorsalFile")
def test_get_latest_file_annotation_invalid_mode(mock_dorsal_file_cls, mock_shared_client):
    """Test that an invalid mode raises a ValueError."""
    mock_instance = mock_dorsal_file_cls.return_value

    mock_instance.get_latest_annotation.return_value = GenericFileAnnotation()

    with pytest.raises(ValueError, match="Invalid mode: 'csv'"):
        file_api.get_latest_file_annotation("hash_abc", "open/schema", mode="csv")  # type: ignore


def test_list_file_annotations_invalid_mode():
    """Test that an invalid mode raises a ValueError."""
    with pytest.raises(ValueError, match="Invalid mode: 'xml'"):
        file_api.list_file_annotations("abc", mode="xml")  # type: ignore


@patch("dorsal.api.file.get_dorsal_file_record")
def test_list_file_annotations_pydantic(mock_get_record):
    """Test retrieving annotations as a pydantic object/dictionary directly."""
    mock_record = MagicMock()
    mock_record.annotations = {"AudioTranscription": [{"id": "123"}]}
    mock_get_record.return_value = mock_record

    result = file_api.list_file_annotations("hash_123", mode="pydantic")

    mock_get_record.assert_called_once_with(hash_string="hash_123", mode="pydantic", api_key=None)
    assert result == {"AudioTranscription": [{"id": "123"}]}


@patch("dorsal.api.file.get_dorsal_file_record")
def test_list_file_annotations_dict(mock_get_record):
    """Test retrieving annotations as a standard dictionary via model_dump."""
    mock_record = MagicMock()

    mock_record.model_dump.return_value = {"annotations": {"AudioTranscription": [{"id": "123"}]}}
    mock_get_record.return_value = mock_record

    result = file_api.list_file_annotations("hash_123", mode="dict")
    assert result == {"AudioTranscription": [{"id": "123"}]}


@patch("dorsal.api.file.get_dorsal_file_record")
def test_list_file_annotations_json(mock_get_record):
    """Test retrieving annotations as a JSON string."""
    mock_record = MagicMock()
    mock_record.model_dump.return_value = {"annotations": {"AudioTranscription": [{"id": "123"}]}}
    mock_get_record.return_value = mock_record

    result = file_api.list_file_annotations("hash_123", mode="json")
    assert isinstance(result, str)
    assert '"AudioTranscription"' in result
    assert '"123"' in result


@patch("dorsal.api.file.get_dorsal_file_record")
def test_list_file_annotations_exception(mock_get_record):
    """Test that unexpected exceptions are logged and re-raised."""
    mock_get_record.side_effect = Exception("API Offline")

    with pytest.raises(Exception, match="API Offline"):
        file_api.list_file_annotations("hash_123")



# Copyright 2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import json
import pathlib
from typing import Annotated, Optional

import typer

logger = logging.getLogger(__name__)


def export_adapter(
    ctx: typer.Context,
    file_path: Annotated[
        pathlib.Path,
        typer.Argument(
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            help="Path to the Dorsal JSON output or raw record JSON.",
        ),
    ],
    target_format: Annotated[
        Optional[str],
        typer.Argument(
            help="The target format to export to (e.g., 'srt', 'vtt'). Optional if inferred from --output.",
        ),
    ] = None,
    schema_id: Annotated[
        Optional[str],
        typer.Option(
            "--schema-id",
            "-s",
            help="Explicitly pass the schema ID. Required if passing a raw record without a wrapper.",
        ),
    ] = None,
    output_path: Annotated[
        Optional[pathlib.Path],
        typer.Option(
            "--output",
            "-o",
            help="Custom output file/directory path. Overrides auto-save naming.",
        ),
    ] = None,
    no_save: Annotated[
        bool,
        typer.Option(
            "--no-save",
            help="Disable auto-saving results to disk (useful for read-only environments or pure piping).",
        ),
    ] = False,
):
    """
    Export a Dorsal JSON result into a different format using an adapter.
    """
    from dorsal.common.cli import exit_cli, EXIT_CODE_ERROR, get_rich_console, get_error_console
    from dorsal.cli.adapter_app.helpers import extract_records
    from dorsal.api.adapters import export_record, export_record_to_file, get_format_extension
    from rich.table import Table

    console = get_rich_console()
    error_console = get_error_console()
    palette: dict[str, str] = ctx.obj.get("palette", {}) if ctx.obj else {}

    try:
        data = json.loads(file_path.read_text(encoding="utf-8"))
    except json.JSONDecodeError as e:
        error_console.print(f"[{palette.get('error', 'bold red')}]Invalid JSON file:[/] {e}")
        exit_cli(code=EXIT_CODE_ERROR)

    try:
        records_to_process = extract_records(data, schema_override=schema_id)
    except ValueError as e:
        error_console.print(f"[{palette.get('error', 'bold red')}]Validation Error:[/] {e}")
        exit_cli(code=EXIT_CODE_ERROR)

    resolved_format = target_format
    if not resolved_format:
        if output_path and not output_path.is_dir() and output_path.suffix:
            resolved_format = output_path.suffix.lstrip(".")
        else:
            error_console.print(
                f"[{palette.get('error', 'bold red')}]Usage Error:[/] You must provide a "
                "'target_format' argument (e.g., 'srt') unless an '--output' file path "
                "with an extension is provided."
            )
            exit_cli(code=EXIT_CODE_ERROR)

    is_batch = len(records_to_process) > 1
    terminal_outputs = []
    results_data = []
    saved_paths_msg = []

    out_dir = output_path if output_path and output_path.is_dir() else pathlib.Path.cwd()

    for i, (current_schema_id, record, orig_file_path) in enumerate(records_to_process):
        if orig_file_path:
            input_name = pathlib.Path(orig_file_path).name
            base_name = pathlib.Path(orig_file_path).stem
        else:
            input_name = f"Record {i + 1}"
            base_name = (
                file_path.name.replace(".dorsal.json", "")
                if file_path.name.endswith(".dorsal.json")
                else file_path.stem
            )
            if is_batch:
                base_name = f"{base_name}_{i + 1}"

        if output_path and not output_path.is_dir() and not is_batch:
            save_path = output_path
        else:
            ext = get_format_extension(current_schema_id, resolved_format)
            save_path = out_dir / f"{base_name}.{ext}"

        try:
            if no_save:
                exported_text = export_record(record, current_schema_id, target_format=resolved_format)
            else:
                exported_text = export_record_to_file(
                    record=record, output_path=save_path, schema_id=current_schema_id, target_format=resolved_format
                )
                saved_paths_msg.append(str(save_path.resolve()))

            terminal_outputs.append(exported_text)
            results_data.append(
                {"input_name": input_name, "status": "Success", "output_name": save_path.name if not no_save else "-"}
            )
        except Exception as e:
            logger.exception("Adapter export failed.")
            results_data.append({"input_name": input_name, "status": "Error", "error": str(e)})
            if not is_batch:
                error_console.print(
                    f"[{palette.get('error', 'bold red')}]Export Failed for schema '{current_schema_id}':[/] {e}"
                )
                exit_cli(code=EXIT_CODE_ERROR)

    if not is_batch:
        console.print(terminal_outputs[0], end="")
        if not no_save and saved_paths_msg:
            error_console.print(
                f"\n[{palette.get('info', 'dim')}]Outputs saved successfully:\n  ↳ {saved_paths_msg[0]}[/]"
            )
    else:
        table = Table(
            title=f"[{palette.get('panel_title', 'bold')}]Export Results: {file_path.name}[/]",
            header_style=palette.get("table_header", "bold blue"),
        )
        table.add_column("Input File")
        table.add_column("Status")

        if not no_save:
            table.add_column(f"{resolved_format.upper()} Output")

        for item in results_data:
            row = []
            if item["status"] == "Error":
                status_text = f"[{palette.get('error', 'bold red')}]Error: {item.get('error')}[/]"
                row.extend([item["input_name"], status_text])
                if not no_save:
                    row.append("-")
            else:
                status_text = f"[{palette.get('success', 'green')}]Success[/]"
                row.extend([item["input_name"], status_text])
                if not no_save:
                    row.append(f"[{palette.get('primary_value', 'cyan')}]{item['output_name']}[/]")

            table.add_row(*row)

        console.print(table)

        if not no_save and saved_paths_msg:
            error_console.print(f"\n[{palette.get('info', 'dim')}]Complete. Files saved to {out_dir.resolve()}[/]")



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import pytest
from unittest.mock import patch, MagicMock

from dorsal.common import environment


def test_is_jupyter_environment_true():
    """Test the case where the code IS running in a Jupyter kernel."""
    # Create a mock for the IPython module and its get_ipython function
    mock_ipython = MagicMock()
    mock_ipython.get_ipython.return_value.config = {"IPKernelApp": {}}

    # Temporarily inject our mock module into the system's list of modules
    with patch.dict(sys.modules, {"IPython": mock_ipython}):
        assert environment.is_jupyter_environment() is True


def test_is_jupyter_environment_false_not_kernel():
    """Test the case where IPython is present but it's not a Jupyter kernel."""
    mock_ipython = MagicMock()
    mock_ipython.get_ipython.return_value.config = {}  # Config does not contain the key

    with patch.dict(sys.modules, {"IPython": mock_ipython}):
        assert environment.is_jupyter_environment() is False


def test_is_jupyter_environment_no_ipython_module():
    """Test the case where the IPython module is not installed, causing an ImportError."""
    # Temporarily remove IPython from sys.modules to simulate it not being installed
    with patch.dict(sys.modules, {"IPython": None}):
        assert environment.is_jupyter_environment() is False


def test_is_jupyter_environment_ipython_returns_none():
    """Test the case where get_ipython() returns None, causing an AttributeError."""
    mock_ipython = MagicMock()
    mock_ipython.get_ipython.return_value = None  # Simulate being in a non-interactive session

    with patch.dict(sys.modules, {"IPython": mock_ipython}):
        assert environment.is_jupyter_environment() is False


def test_is_jupyter_environment_generic_exception():
    """Test that any other unexpected exception is caught and returns False."""
    mock_ipython = MagicMock()
    mock_ipython.get_ipython.side_effect = RuntimeError("A generic, unexpected error")

    with patch.dict(sys.modules, {"IPython": mock_ipython}):
        assert environment.is_jupyter_environment() is False



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import platform
from pathlib import Path


def initialize_mediainfo():
    """Checks for MediaInfo availability and returns the library and its path if needed."""
    try:
        from pymediainfo import MediaInfo  # type: ignore

        library_path = None
        # On macos, `homebrew` may install to a non-standard path that pymediainfo may not search by default
        # See: https://docs.brew.sh/Installation#alternative-installs
        # See: https://github.com/sbraz/pymediainfo/issues/151
        if sys.platform == "darwin" and platform.machine() == "arm64":
            homebrew_path = Path("/opt/homebrew/lib/libmediainfo.dylib")
            if homebrew_path.is_file():
                library_path = str(homebrew_path)

        MediaInfo._get_library(library_file=library_path)

        return MediaInfo, library_path

    except (ImportError, OSError) as e:
        if "failed to load library" in str(e).lower() or isinstance(e, ImportError):
            if sys.platform == "darwin":
                message = (
                    "The 'mediainfo' library is not installed or could not be found. "
                    "On macOS, it can be installed with Homebrew:\n\n"
                    "    brew install mediainfo\n"
                )
            elif sys.platform.startswith("linux"):
                message = (
                    "The 'mediainfo' library is not installed. "
                    "Please install it using your system's package manager, for example:\n\n"
                    "    On Debian/Ubuntu: sudo apt-get install mediainfo\n"
                    "    On Fedora/CentOS: sudo dnf install mediainfo\n"
                )
            else:
                message = f"The 'mediainfo' library could not be found on your system ({sys.platform}). Please install it to continue."

            raise ImportError(message) from e
        else:
            raise


def initialize_magic():
    """
    Checks for python-magic availability and returns the library and Exception class.
    """
    try:
        import magic
        from magic import MagicException

        return magic, MagicException
    except ImportError as err:
        if "failed to find libmagic" in str(err).lower() or "magic library not found" in str(err).lower():
            if sys.platform == "darwin":
                message = (
                    "The 'libmagic' library is not installed. "
                    "Please install it using Homebrew:\n\n"
                    "    brew install libmagic\n"
                )
            elif sys.platform.startswith("linux"):
                message = (
                    "The 'libmagic' library is not installed. "
                    "Please install it using your system's package manager, for example:\n\n"
                    "    On Debian/Ubuntu: sudo apt-get install libmagic1\n"
                    "    On Fedora/CentOS: sudo dnf install file-libs\n"
                )
            else:
                message = f"The 'libmagic' library could not be found on your system ({sys.platform}). Please install it to continue."
            raise ImportError(message) from err
        else:
            raise err



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typer
from typing import Annotated
import pathlib
import logging
from rich.panel import Panel
from rich.text import Text
import json

logger = logging.getLogger(__name__)


def push_file(
    ctx: typer.Context,
    path: Annotated[
        pathlib.Path,
        typer.Argument(
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            help="The path to the single file to push to DorsalHub.",
        ),
    ],
    public: Annotated[
        bool,
        typer.Option("--public/--private", help="Index the file record as public or private."),
    ] = False,
    use_cache: Annotated[
        bool,
        typer.Option(
            "--use-cache",
            help="Force the use of the cache, overriding any global setting.",
            rich_help_panel="Cache Options",
        ),
    ] = False,
    skip_cache: Annotated[
        bool,
        typer.Option(
            "--skip-cache",
            help="Bypass the local cache and re-process the file (does not update cache).",
            rich_help_panel="Cache Options",
        ),
    ] = False,
    overwrite_cache: Annotated[
        bool,
        typer.Option(
            "--overwrite-cache",
            help="Re-process the file and overwrite the local cache with new data.",
            rich_help_panel="Cache Options",
        ),
    ] = False,
    strict: Annotated[
        bool,
        typer.Option(
            "--strict",
            help="Fail with an error code if any metadata is rejected (Partial Success).",
            rich_help_panel="Integrity Options",
        ),
    ] = False,
    json_output: Annotated[
        bool,
        typer.Option("--json", help="Output the API response as a raw JSON object."),
    ] = False,
    resolve_links: Annotated[
        bool,
        typer.Option(
            "--follow-links/--no-follow-links",
            help="Follow symlinks to index target metadata vs indexing the link itself.",
        ),
    ] = True,
):
    """
    Pushes a single file's metadata to DorsalHub.
    """
    from dorsal.common.cli import (
        EXIT_CODE_ERROR,
        get_rich_console,
        determine_use_cache_value,
        exit_cli,
    )
    from dorsal.file.dorsal_file import LocalFile
    from dorsal.common.exceptions import (
        DorsalClientError,
        DorsalOfflineError,
        AuthError,
        PartialIndexingError,
    )

    console = get_rich_console()

    if use_cache and skip_cache:
        exit_cli(
            code=EXIT_CODE_ERROR,
            message="Error: --use-cache and --skip-cache flags cannot be used together.",
        )

    if skip_cache and overwrite_cache:
        exit_cli(
            code=EXIT_CODE_ERROR,
            message="Error: --skip-cache and --overwrite-cache flags cannot be used together.",
        )

    use_cache_value = determine_use_cache_value(use_cache=use_cache, skip_cache=skip_cache)
    palette: dict[str, str] = ctx.obj["palette"]
    access_level_str = "public" if public else "private"

    if not json_output:
        console.print(
            f"📡 Preparing to push metadata for [{palette['primary_value']}]{path.name}[/] as a {access_level_str} record..."
        )

    try:
        local_file = LocalFile(
            file_path=str(path),
            use_cache=use_cache_value,
            overwrite_cache=overwrite_cache,
            follow_symlinks=resolve_links,
        )

        logger.debug("Record to push: %s", local_file.model_dump_json(exclude_none=True, by_alias=True))

        with console.status("Pushing to DorsalHub..."):
            api_response = local_file.push(public=public, strict=strict)

        if json_output:
            console.print(json.dumps(api_response.model_dump(mode="json"), indent=2, ensure_ascii=False))
            exit_cli()

        if api_response.results and api_response.success > 0:
            pushed_hash = api_response.results[0].hash
            success_text = Text.assemble(
                (
                    "The file record was successfully pushed to DorsalHub.\n\n",
                    palette.get("success", "bold green"),
                ),
                ("SHA256 Hash: ", palette.get("key", "dim")),
                (f"{pushed_hash}", palette.get("hash_value", "magenta")),
            )
            panel_title = "✅ Push Complete"
            panel_border_style = palette.get("panel_border_success", "green")
        else:
            detail: str | None = "Unknown"
            if api_response.results and hasattr(api_response.results[0], "detail"):
                if api_response.results[0].annotations:
                    detail = api_response.results[0].annotations[0].detail

            success_text = Text(
                f"The file could not be pushed to DorsalHub.\nReason: {detail}",
                style=palette.get("error", "bold red"),
            )
            panel_title = "❌ Push Failed"
            panel_border_style = palette.get("panel_border_error", "red")

        console.print(
            Panel(
                success_text,
                expand=False,
                title=panel_title,
                border_style=panel_border_style,
            )
        )

    except PartialIndexingError as e:
        if json_output:
            error_payload = {
                "error": "PartialIndexingError",
                "message": str(e),
                "summary": e.summary,
            }
            console.print(json.dumps(error_payload, indent=2, ensure_ascii=False))
        else:
            console.print(f"[{palette.get('error', 'bold red')}]Strict Mode Error:[/] {e}")
            if e.summary and "failures" in e.summary:
                console.print(f"[{palette.get('warning', 'yellow')}]Failures detected:[/]")
                for failure in e.summary["failures"]:
                    console.print(f"  - {failure}")

        exit_cli(code=EXIT_CODE_ERROR)

    except typer.Exit:
        raise
    except DorsalOfflineError:
        raise
    except AuthError:
        raise
    except DorsalClientError as e:
        exit_cli(code=EXIT_CODE_ERROR, message=f"API Error: {e.message}")
    except Exception as e:
        logger.exception(f"An unexpected error occurred while pushing file {path}.")
        exit_cli(code=EXIT_CODE_ERROR, message=f"An unexpected error occurred: {e}")



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import json
import datetime
from unittest.mock import MagicMock, patch, mock_open

from dorsal.file.collection.local import LocalFileCollection
from dorsal.common.exceptions import InvalidTagError, DorsalError, SyncConflictError
from dorsal.file.dorsal_file import LocalFile
from dorsal.file.validators.file_record import FileRecordStrict, NewFileTag, ValidateTagsResult
from dorsal.client.validators import FileIndexResponse


@pytest.fixture
def mock_metadata_reader():
    """Mocks the MetadataReader to avoid filesystem scans."""
    with patch("dorsal.file.collection.local.MetadataReader") as mock_reader_class:
        mock_instance = MagicMock()
        mock_reader_class.return_value = mock_instance
        yield mock_instance


# --- Tests ---


def test_local_collection_init_from_path(mock_metadata_reader):
    """Test initialization from a directory path string."""
    mock_file = MagicMock(spec=LocalFile)
    mock_metadata_reader.scan_directory.return_value = ([mock_file], ["warning1"])
    collection = LocalFileCollection(source="/fake/dir")
    mock_metadata_reader.scan_directory.assert_called_once_with(
        dir_path="/fake/dir",
        recursive=False,
        return_errors=True,
        console=None,
        palette=None,
        skip_cache=False,
        overwrite_cache=False,
        follow_symlinks=True,
        lazy=False,
    )
    assert len(collection) == 1
    assert collection.warnings == ["warning1"]


def test_info_method():
    """Test the info() method for accurate statistical summaries."""
    # Arrange
    now = datetime.datetime.now(tz=datetime.UTC)
    file1 = MagicMock(
        spec=LocalFile,
        size=2000,
        media_type="application/pdf",
        _source="disk",
        date_modified=now,
    )
    file1.name = "f1.pdf"
    file2 = MagicMock(
        spec=LocalFile,
        size=1000,
        media_type="image/jpeg",
        _source="cache",
        date_modified=now,
    )
    file2.name = "f2.jpg"
    file3 = MagicMock(
        spec=LocalFile,
        size=3000,
        media_type="application/pdf",
        _source="disk",
        date_modified=now,
    )
    file3.name = "f3.pdf"
    collection = LocalFileCollection(source=[file1, file2, file3])

    info = collection.info()

    assert "overall" in info
    assert info["overall"]["total_files"] == 3
    assert info["overall"]["total_size"] == 6000
    assert info["overall"]["smallest_file"]["path"] == "f2.jpg"
    assert info["overall"]["largest_file"]["path"] == "f3.pdf"


def test_find_duplicates():
    """Test the find_duplicates() method."""

    file1 = MagicMock(spec=LocalFile, hash="duplicate", size=150)
    file1.name = "a.txt"
    file1.to_dict.return_value = {"name": "a.txt"}
    file2 = MagicMock(spec=LocalFile, hash="unique", size=200)
    file2.name = "b.txt"
    file3 = MagicMock(spec=LocalFile, hash="duplicate", size=150)
    file3.name = "c.txt"
    file3.to_dict.return_value = {"name": "c.txt"}
    collection = LocalFileCollection(source=[file1, file2, file3])

    result = collection.find_duplicates()

    assert result["total_sets"] == 1
    assert result["total_wasted_space"] == "150 B"


def test_filter_method():
    """Test the filter() method with various conditions."""
    # Arrange
    file1 = MagicMock(spec=LocalFile, media_type="image/jpeg", size=100)
    file2 = MagicMock(spec=LocalFile, media_type="image/jpeg", size=1000)
    file3 = MagicMock(spec=LocalFile, media_type="text/plain", size=500)
    collection = LocalFileCollection(source=[file1, file2, file3])

    filtered_by_type = collection.filter(media_type="image/jpeg")
    filtered_by_size = collection.filter(size__gt=600)

    assert len(filtered_by_type) == 2
    assert len(filtered_by_size) == 1
    assert filtered_by_size.files[0].size == 1000


@patch("dorsal.file.collection.local.get_shared_dorsal_client")
def test_add_tags(mock_get_client):
    """Test successfully adding tags to all files in a collection."""
    # Arrange
    mock_client = MagicMock()
    mock_client.validate_tag.return_value = ValidateTagsResult(valid=True)
    mock_get_client.return_value = mock_client

    file1 = MagicMock(spec=LocalFile)
    file1._add_local_tag = MagicMock()
    file2 = MagicMock(spec=LocalFile)
    file2._add_local_tag = MagicMock()
    collection = LocalFileCollection(source=[file1, file2])
    tags_to_add = [{"name": "status", "value": "approved"}]

    collection.add_tags(tags=tags_to_add)

    file1._add_local_tag.assert_called_once_with(name="status", value="approved")


def test_to_json_export():
    """Test exporting the collection to a JSON string."""
    file1 = MagicMock(spec=LocalFile)
    file1.to_dict.return_value = {"name": "a.txt", "hash": "h1"}
    file1.date_modified = datetime.datetime(2025, 1, 1, 1, 1, tzinfo=datetime.UTC)
    file1.date_created = datetime.datetime(2025, 1, 1, 1, 1, tzinfo=datetime.UTC)
    file1.size = 1000
    file1.name = "a.txt"
    file1.media_type = "text/plain"
    file1._source = "test"
    file1._file_path = "a.txt"

    collection = LocalFileCollection(source=[file1], source_info={"path": "/fake/dir"})

    assert collection._is_populated is True

    json_output = collection.to_json()
    data = json.loads(json_output)

    assert data["scan_metadata"]["path"] == "/fake/dir"
    assert len(data["results"]) == 1
    assert data["results"][0]["name"] == "a.txt"


@patch("builtins.open", new_callable=mock_open)
def test_to_csv_export(mock_file_open):
    """Test exporting the collection to a CSV file."""
    file1 = MagicMock(spec=LocalFile, hash="h1", _file_path="/fake/a.txt")
    collection = LocalFileCollection(source=[file1], source_info={"path": "/fake/"})

    collection.to_csv("output.csv")

    mock_file_open.assert_called_once_with(file="output.csv", mode="w", newline="", encoding="utf-8")
    handle = mock_file_open()
    written_data = "".join(call.args[0] for call in handle.write.call_args_list)

    assert "hash,file_path,source_path" in written_data
    assert "h1,/fake/a.txt,/fake/" in written_data


@patch("dorsal.file.collection.local.get_shared_dorsal_client")
def test_push(mock_get_client):
    """Test pushing file records to the API."""
    mock_client = MagicMock()

    mock_response = MagicMock(spec=FileIndexResponse)
    mock_response.success = 1
    mock_response.results = []
    mock_client.index_private_file_records.return_value = mock_response

    mock_get_client.return_value = mock_client

    file1 = MagicMock(spec=LocalFile)
    file1.validation_hash = "b" * 64
    file1.model = MagicMock(spec=FileRecordStrict)
    collection = LocalFileCollection(source=[file1])
    collection._client = mock_client

    summary = collection.push(public=False)

    mock_client.index_private_file_records.assert_called_once_with(file_records=[file1.model])
    assert summary["success"] == 1


@patch("dorsal.file.collection.remote.DorsalFileCollection")
@patch("dorsal.file.collection.local.get_shared_dorsal_client")
def test_create_remote_collection(mock_get_client, mock_remote_class):
    """Test the multi-step process of creating a new remote collection."""
    mock_client = MagicMock()
    mock_get_client.return_value = mock_client

    mock_client.create_collection.return_value = MagicMock(collection_id="new_col_123")
    mock_client.add_files_to_collection.return_value = MagicMock(added_count=1, duplicate_count=0)
    final_state = MagicMock()
    final_state.collection.date_modified = datetime.datetime.now()
    final_state.collection.file_count = 1
    mock_client.get_collection.return_value = final_state

    file1 = MagicMock(spec=LocalFile, hash="h1", validation_hash="vh1")
    file1.model = MagicMock()
    collection = LocalFileCollection(source=[file1])
    collection._client = mock_client

    with patch.object(collection, "push", return_value={"success": 1}) as mock_push:
        # Act
        collection.create_remote_collection(name="New Test Collection")

        # Assert
        mock_push.assert_called_once_with(public=False, api_key=None)
        mock_client.create_collection.assert_called_once()
        assert collection.remote_collection_id == "new_col_123"


def test_to_sqlite_export(tmp_path):
    """Test exporting the collection to an SQLite database by inspecting the output file."""
    sqlite3 = pytest.importorskip("sqlite3")

    file1 = MagicMock(spec=LocalFile, hash="h1", _file_path="/fake/a.txt")
    collection = LocalFileCollection(source=[file1], source_info={"path": "/fake/"})

    db_path = tmp_path / "test.db"

    collection.to_sqlite(str(db_path))

    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()
    cursor.execute("SELECT hash, file_path, source_path FROM files")
    result = cursor.fetchone()
    conn.close()

    assert result is not None
    assert result[0] == "h1"
    assert result[1] == "/fake/a.txt"
    assert result[2] == "/fake/"


def test_to_dataframe_export():
    """Test exporting the collection to a pandas DataFrame by inspecting the output."""
    pd = pytest.importorskip("pandas")

    file1 = MagicMock(spec=LocalFile, hash="h1", name="a.txt", _file_path="/fake/a.txt")
    collection = LocalFileCollection(source=[file1], source_info={"path": "/fake/"})

    df = collection.to_dataframe()

    assert isinstance(df, pd.DataFrame)
    assert len(df) == 1
    assert list(df.columns) == ["hash", "file_path", "source_path"]
    assert df.iloc[0]["hash"] == "h1"
    assert df.iloc[0]["file_path"] == "/fake/a.txt"


@patch("dorsal.file.collection.local.is_permitted_public_media_type")
def test_push_public_raises_error_for_restricted_types(mock_is_permitted):
    """
    Test that pushing with public=True raises a ValueError if the collection
    contains files with restricted media types.
    """
    # Arrange: Force the permission check to fail
    mock_is_permitted.return_value = False

    file1 = MagicMock(spec=LocalFile, media_type="application/secret", hash="h1")
    file1.name = "secret_plans.doc"

    collection = LocalFileCollection(source=[file1])

    # Act & Assert
    with pytest.raises(ValueError) as exc_info:
        collection.push(public=True)

    error_msg = str(exc_info.value)
    assert "restricted media types" in error_msg
    assert "'secret_plans.doc' (application/secret)" in error_msg


def test_collection_iteration_and_access_types():
    file1 = MagicMock(spec=LocalFile)
    file1.name = "file1"
    file1._file_path = "/local/path/file1"

    collection = LocalFileCollection(files=[file1])

    items = list(collection)
    assert len(items) == 1
    assert items[0] is file1
    assert items[0]._file_path == "/local/path/file1"

    item = collection[0]
    assert item is file1
    assert item._file_path == "/local/path/file1"



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import json
from typer.testing import CliRunner
from unittest.mock import MagicMock
from rich.panel import Panel

from dorsal.cli import app
from dorsal.common.exceptions import NotFoundError, DuplicateTagError, DorsalClientError

runner = CliRunner()
HASH_STRING = "sha256:123abc456def"
TAG_ID = "tag_xyz789"


@pytest.fixture
def mock_tag_app_cmds(mocker):
    """Mocks dependencies for the `record tag` subcommands."""
    # Mock response object
    mock_add_response = MagicMock()
    mock_add_response.model_dump_json.return_value = '{"success": true}'

    # Mock standard tag addition
    mock_add = mocker.patch("dorsal.api.file.add_tag_to_file", return_value=mock_add_response)

    mock_add_label = mocker.patch("dorsal.api.file.add_label_to_file", return_value=mock_add_response)

    # Mock removal
    mock_remove = mocker.patch("dorsal.api.file.remove_tag_from_file")

    return {
        "add_tag": mock_add,
        "add_label": mock_add_label,
        "add_response": mock_add_response,
        "remove_tag": mock_remove,
    }


# --- Tests for `record tag add` ---


def test_add_tag_success_private(mock_rich_console, mock_tag_app_cmds):
    """Tests adding a private tag (the default)."""
    result = runner.invoke(app, ["record", "tag", "add", HASH_STRING, "--name", "key", "--value", "val"])

    assert result.exit_code == 0
    mock_tag_app_cmds["add_tag"].assert_called_once_with(hash_string=HASH_STRING, name="key", value="val", public=False)
    assert "Successfully added tag" in mock_rich_console.print.call_args.args[0]


def test_add_tag_success_public(mock_rich_console, mock_tag_app_cmds):
    """Tests adding a public tag with the --public flag."""
    result = runner.invoke(
        app,
        [
            "record",
            "tag",
            "add",
            HASH_STRING,
            "--name",
            "key",
            "--value",
            "val",
            "--public",
        ],
    )

    assert result.exit_code == 0
    mock_tag_app_cmds["add_tag"].assert_called_once_with(hash_string=HASH_STRING, name="key", value="val", public=True)


def test_add_label_success(mock_rich_console, mock_tag_app_cmds):
    """Tests adding a simple label using the shorthand argument."""
    # Command: dorsal record tag add <hash> "urgent"
    result = runner.invoke(app, ["record", "tag", "add", HASH_STRING, "urgent"])

    assert result.exit_code == 0
    # Verify it delegates to the specific label API function
    mock_tag_app_cmds["add_label"].assert_called_once_with(hash_string=HASH_STRING, label="urgent")
    # Verify the standard add_tag was NOT called
    mock_tag_app_cmds["add_tag"].assert_not_called()
    assert "(label)" in mock_rich_console.print.call_args_list[0].args[0]


def test_add_label_error_public(mock_rich_console, mock_tag_app_cmds):
    """Tests that combining a label with --public raises an error."""
    # Command: dorsal record tag add <hash> "urgent" --public
    result = runner.invoke(app, ["record", "tag", "add", HASH_STRING, "urgent", "--public"])

    assert result.exit_code == 1

    # Verify API was NOT called
    mock_tag_app_cmds["add_label"].assert_not_called()
    mock_tag_app_cmds["add_tag"].assert_not_called()

    # Verify error message
    printed_object = mock_rich_console.print.call_args.args[0]
    assert isinstance(printed_object, Panel)
    assert "Simple labels must be PRIVATE" in str(printed_object.renderable)


def test_add_label_error_ambiguous(mock_rich_console, mock_tag_app_cmds):
    """Tests that combining a label with --name/--value raises an error."""
    # Command: dorsal record tag add <hash> "urgent" --name "genre"
    result = runner.invoke(app, ["record", "tag", "add", HASH_STRING, "urgent", "--name", "genre"])

    assert result.exit_code == 1
    mock_tag_app_cmds["add_label"].assert_not_called()

    printed_object = mock_rich_console.print.call_args.args[0]
    assert "Ambiguous Request" in str(printed_object.renderable)


def test_add_tag_missing_args(mock_rich_console, mock_tag_app_cmds):
    """Tests error when neither a label nor name/value pairs are provided."""
    # Command: dorsal record tag add <hash>
    result = runner.invoke(app, ["record", "tag", "add", HASH_STRING])

    assert result.exit_code == 1
    printed_object = mock_rich_console.print.call_args.args[0]
    assert "Missing Arguments" in str(printed_object.renderable)


def test_add_tag_json_output(mock_rich_console, mock_tag_app_cmds):
    """Tests the --json output for the add command."""
    result = runner.invoke(
        app,
        ["record", "tag", "add", HASH_STRING, "--name", "k", "--value", "v", "--json"],
    )

    assert result.exit_code == 0
    mock_tag_app_cmds["add_response"].model_dump_json.assert_called_once()
    mock_rich_console.print.assert_called_once_with('{"success": true}')


def test_add_tag_not_found_error(mock_rich_console, mock_tag_app_cmds):
    """Tests error handling when the target file for adding a tag is not found."""
    mock_tag_app_cmds["add_tag"].side_effect = NotFoundError("test")

    result = runner.invoke(app, ["record", "tag", "add", HASH_STRING, "--name", "k", "--value", "v"])

    assert result.exit_code != 0
    printed_object = mock_rich_console.print.call_args.args[0]
    assert isinstance(printed_object, Panel)
    assert "Cannot add tag: No file record found" in str(printed_object.renderable)


def test_remove_tag_success(mock_rich_console, mock_tag_app_cmds):
    """Tests successfully removing a tag."""
    result = runner.invoke(app, ["record", "tag", "rm", HASH_STRING, "--tag-id", TAG_ID])

    assert result.exit_code == 0
    mock_tag_app_cmds["remove_tag"].assert_called_once_with(hash_string=HASH_STRING, tag_id=TAG_ID)
    assert "Successfully removed tag" in mock_rich_console.print.call_args.args[0]


def test_remove_tag_json_output(mock_rich_console, mock_tag_app_cmds):
    """Tests the --json output for the remove command."""
    result = runner.invoke(app, ["record", "tag", "rm", HASH_STRING, "--tag-id", TAG_ID, "--json"])

    assert result.exit_code == 0
    mock_rich_console.print.assert_called_once()
    json_str = mock_rich_console.print.call_args.args[0]
    data = json.loads(json_str)
    assert data["success"] is True
    assert f"Tag '{TAG_ID}' removed" in data["detail"]


def test_remove_tag_not_found_error(mock_rich_console, mock_tag_app_cmds):
    """Tests error handling when the file or tag to remove is not found."""
    mock_tag_app_cmds["remove_tag"].side_effect = NotFoundError("test")

    result = runner.invoke(app, ["record", "tag", "rm", HASH_STRING, "--tag-id", TAG_ID])

    assert result.exit_code != 0
    printed_object = mock_rich_console.print.call_args.args[0]
    assert isinstance(printed_object, Panel)
    assert "Could not find a file with that hash" in str(printed_object.renderable)



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from typing import Any

from dorsal.common.model import AnnotationModel

from dorsal.file.annotation_models.ebook.config import EBOOK_FORMAT_MAPPING
import dorsal.file.annotation_models.ebook.utils as ebook_utils

logger = logging.getLogger(__name__)


class EbookAnnotationModel(AnnotationModel):
    """Extracts metadata from common ebook formats (currently only supports Epub)."""

    id: str = "dorsal/ebook"
    version: str = "0.1.0"
    variant: str = "dispatcher"

    def main(self) -> dict[str, Any] | None:
        """
        Extracts metadata by dispatching to the correct format-specific parser.

        Returns:
          * Dictionary of ebook metadata if successful.
          * None if the format is unsupported or parsing fails.
        """
        logger.debug(
            "EbookAnnotationModel: Starting metadata extraction for '%s'",
            self.file_path,
        )

        try:
            _, ext = os.path.splitext(self.file_path)
            parser_type = EBOOK_FORMAT_MAPPING.get(ext.lower())

            metadata: dict[str, Any] | None = None

            if parser_type == "epub":
                self.variant = "epub_stdlib"
                metadata = ebook_utils.extract_epub_metadata(self.file_path)

            else:
                self.error = f"Unsupported ebook format: '{ext}' for file: {self.file_path}"
                logger.info(self.error)
                return None

            if metadata is None:
                self.error = f"Failed to parse ebook metadata for file: {self.file_path} (parser: {self.variant})"
                logger.warning(self.error)
                return None

            logger.debug(
                "EbookAnnotationModel: Successfully processed '%s' with parser '%s'",
                self.file_path,
                self.variant,
            )
            return metadata

        except ImportError as e:
            self.error = f"Missing dependency for parser '{self.variant}': {e}. Cannot process file."
            logger.error(self.error, exc_info=True)
            return None

        except (FileNotFoundError, IOError, OSError) as e:
            self.error = f"File system error during ebook processing: {type(e).__name__}: {e}"
            logger.error(
                "EbookAnnotationModel: CRITICAL OS/IO Error for '%s'. Error: %s",
                self.file_path,
                self.error,
                exc_info=True,
            )
            raise

        except Exception as e:
            self.error = f"Unexpected error during EbookAnnotationModel processing: {type(e).__name__}: {e}"
            logger.error(
                "EbookAnnotationModel: UNEXPECTED Error for '%s'. Error: %s",
                self.file_path,
                self.error,
                exc_info=True,
            )
            raise



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import pytest
from unittest.mock import MagicMock
from dorsal.file.utils.reports import (
    get_collection_overview_data,
    get_dynamic_size_histogram_data,
    resolve_template_path,
    TemplateNotFoundError,
)


# Mocking the LocalFileCollection interface
class MockFile:
    def __init__(self, name, size, media_type, extension, date_modified):
        self.name = name
        self.size = size
        self.media_type = media_type
        self.extension = extension
        self.date_modified = date_modified

    def to_dict(self):
        return {"name": self.name, "size": self.size, "date_modified": self.date_modified.isoformat()}


@pytest.fixture
def mock_collection():
    collection = MagicMock()

    # Create a diverse set of files
    files = [
        MockFile("a.txt", 100, "text/plain", ".txt", datetime.datetime(2023, 1, 1)),
        MockFile("b.jpg", 5000, "image/jpeg", ".jpg", datetime.datetime(2023, 1, 2)),
        MockFile("c.pdf", 200, "application/pdf", ".pdf", datetime.datetime(2023, 1, 3)),
        MockFile("d.txt", 100, "text/plain", ".txt", datetime.datetime(2023, 1, 4)),
    ]

    collection.files = files

    # Mock the .info() return
    collection.info.return_value = {
        "overall": {"total_size": 5400, "count": 4},
        "by_type": [
            {"media_type": "image/jpeg", "total_size": 5000, "count": 1},
            {"media_type": "application/pdf", "total_size": 200, "count": 1},
            {"media_type": "text/plain", "total_size": 200, "count": 2},
        ],
    }
    return collection


def test_get_collection_overview_data(mock_collection):
    """Smoke test for collection overview data generation."""
    data = get_collection_overview_data(mock_collection)

    assert data["most_recent_file_record"]["name"] == "d.txt"

    # Check Media Type Aggregation
    assert len(data["media_type"]["by_count"]) == 3
    text_entry = next(x for x in data["media_type"]["by_count"] if x["media_type"] == "text/plain")
    assert text_entry["count"] == 2

    # Check Extension Aggregation
    assert len(data["extension"]["by_size"]) == 3
    jpg_entry = next(x for x in data["extension"]["by_size"] if x["extension"] == ".jpg")
    assert jpg_entry["total_size"] == 5000

    # Check Timeline Data
    assert len(data["timeline_data"]) == 4
    assert data["timeline_data"][0]["y"] == "a.txt"


def test_get_collection_overview_empty():
    """Test behavior with an empty collection."""
    collection = MagicMock()
    collection.files = []
    data = get_collection_overview_data(collection)
    assert data == {}


def test_get_dynamic_size_histogram_data(mock_collection):
    """Test histogram binning logic."""
    data = get_dynamic_size_histogram_data(mock_collection)

    assert len(data) > 0
    assert "bin_label" in data[0]
    assert "count" in data[0]


def test_get_dynamic_size_histogram_large_dataset():
    """Test histogram binning with enough data to trigger the statistical binning logic."""
    collection = MagicMock()
    # Generate 50 files with varying sizes
    collection.files = [MockFile(f"f{i}", i * 100, "t", ".t", datetime.datetime.now()) for i in range(50)]

    data = get_dynamic_size_histogram_data(collection)
    assert len(data) > 1  # Should have multiple bins
    # Check that it produced valid chart data structure
    assert all("bin_label" in item and "count" in item for item in data)


def test_resolve_template_path_failure():
    """Test that invalid templates raise the correct error."""
    # We don't want to test file system search (too fragile), just the failure case
    with pytest.raises(TemplateNotFoundError):
        resolve_template_path("report_type", "non_existent_template_xyz_123")



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import re
from dorsal.file.dependencies import (
    make_media_type_dependency,
    make_file_extension_dependency,
    make_file_size_dependency,
    make_file_name_dependency,
)
from dorsal.file.configs.model_runner import (
    MediaTypeDependencyConfig,
    FileExtensionDependencyConfig,
    FileSizeDependencyConfig,
    FilenameDependencyConfig,
)


def test_make_media_type_success():
    dep = make_media_type_dependency(include=["application/pdf"], exclude=["text/plain"])
    assert isinstance(dep, MediaTypeDependencyConfig)
    assert dep.include == {"application/pdf"}
    assert dep.exclude == {"text/plain"}


def test_make_media_type_fail_str_input():
    with pytest.raises(TypeError) as exc:
        make_media_type_dependency(include="application/pdf")
    assert "Did you mean: include=" in str(exc.value)

    with pytest.raises(TypeError) as exc:
        make_media_type_dependency(exclude="text/plain")
    assert "Did you mean: exclude=" in str(exc.value)


def test_make_media_type_fail_no_args():
    with pytest.raises(ValueError) as exc:
        make_media_type_dependency()
    assert "must have at least one rule" in str(exc.value)


def test_make_extension_success():
    # Test normalization (adding dot, lowering case)
    dep = make_file_extension_dependency(extensions=["PDF", ".txt"])
    assert isinstance(dep, FileExtensionDependencyConfig)
    assert ".pdf" in dep.extensions
    assert ".txt" in dep.extensions


def test_make_extension_fail_str_input():
    with pytest.raises(TypeError) as exc:
        make_file_extension_dependency(extensions=".pdf")
    assert "must be a sequence" in str(exc.value)


def test_make_extension_fail_empty():
    with pytest.raises(ValueError) as exc:
        make_file_extension_dependency(extensions=[])
    assert "must have at least one extension" in str(exc.value)


# --- File Size Dependency Tests ---


def test_make_size_success():
    dep = make_file_size_dependency(min_size="1KB", max_size=2000)
    assert isinstance(dep, FileSizeDependencyConfig)
    assert dep.min_size == 1000
    assert dep.max_size == 2000


def test_make_size_fail_no_args():
    with pytest.raises(ValueError) as exc:
        make_file_size_dependency()
    assert "must have at least one" in str(exc.value)


# --- File Name Dependency Tests ---


def test_make_filename_success():
    dep = make_file_name_dependency(pattern=r"^data_.*")
    assert isinstance(dep, FilenameDependencyConfig)
    assert dep.pattern == r"^data_.*"


def test_make_filename_fail_empty():
    with pytest.raises(ValueError):
        make_file_name_dependency(pattern="")



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typer
import pathlib
import datetime
from typing_extensions import Annotated, Optional


from dorsal.common.exceptions import DorsalError


def make_file_report(
    ctx: typer.Context,
    path: Annotated[
        pathlib.Path,
        typer.Argument(
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            help="The path to the local file to generate a report for.",
        ),
    ],
    output: Annotated[
        Optional[pathlib.Path],
        typer.Option(
            "--output",
            "-o",
            help="Custom path to save the HTML report. If omitted, a default path in ~/.dorsal/scan/ will be used.",
            dir_okay=True,
            file_okay=True,
            writable=True,
            resolve_path=True,
        ),
    ] = None,
    template: Annotated[
        str,
        typer.Option("--template", "-t", help="Name or path of the report template to use."),
    ] = "default",
    use_cache: Annotated[
        bool,
        typer.Option(
            "--use-cache",
            help="Force the use of the cache, overriding any global setting.",
            rich_help_panel="Cache Options",
        ),
    ] = False,
    skip_cache: Annotated[
        bool,
        typer.Option(
            "--skip-cache",
            help="Bypass the local cache and re-process the file.",
            rich_help_panel="Cache Options",
        ),
    ] = False,
    open_report: Annotated[
        bool,
        typer.Option(
            "--open",
            help="Open the report in the default web browser after generation.",
        ),
    ] = False,
):
    """
    Generates a self-contained, interactive HTML report for a single file.
    """
    from dorsal.api.file import generate_html_file_report
    from dorsal.common import constants
    from dorsal.common.cli import (
        EXIT_CODE_ERROR,
        get_rich_console,
        exit_cli,
        determine_use_cache_value,
    )

    if use_cache and skip_cache:
        exit_cli(
            code=EXIT_CODE_ERROR,
            message="Error: --use-cache and --skip-cache cannot be used together.",
        )

    console = get_rich_console()
    palette = ctx.obj.get("palette", {})

    use_cache_value = determine_use_cache_value(use_cache=use_cache, skip_cache=skip_cache)

    final_output_path: pathlib.Path
    if output:
        final_output_path = output
        if output.is_dir():
            final_output_path = output / f"{path.stem}_report.html"
    else:
        constants.CLI_SCAN_REPORTS_DIR.mkdir(parents=True, exist_ok=True)
        timestamp = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
        safe_filename = path.name.replace(" ", "_")
        final_output_path = constants.CLI_SCAN_REPORTS_DIR / f"{safe_filename}-{timestamp}.html"

    with console.status(f"Generating report for '[bold]{path.name}[/]'..."):
        try:
            generate_html_file_report(
                file_path=str(path),
                output_path=str(final_output_path),
                template=template,
                use_cache=use_cache_value,
            )
        except DorsalError as err:
            exit_cli(code=EXIT_CODE_ERROR, message=f"Failed to generate report: {err}")
        except Exception as err:
            exit_cli(code=EXIT_CODE_ERROR, message=f"An unexpected error occurred: {err}")

    console.print(f"✅ Report saved successfully to: [{palette.get('primary_value')}]{final_output_path}[/]")

    if open_report:
        try:
            import webbrowser

            webbrowser.open(f"file://{final_output_path.resolve()}")
        except Exception as err:
            console.print(f"⚠️  Could not automatically open the report: {err}", style=palette.get("warning", "yellow"))



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pathlib


def get_int_envvar(envvar: str, default: int, min_val: int | None = None, max_val: int | None = None) -> int:
    value_str = os.getenv(envvar)
    if value_str is None:
        return default
    try:
        value = int(value_str)
        if min_val is not None and value < min_val:
            return min_val
        if max_val is not None and value > max_val:
            return max_val
        return value
    except (ValueError, TypeError):
        return default


def get_float_envvar(envvar: str, default: float, min_val: float | None = None, max_val: float | None = None) -> float:
    value_str = os.getenv(envvar)
    if value_str is None:
        return default
    try:
        value = float(value_str)
        if min_val is not None and value < min_val:
            return min_val
        if max_val is not None and value > max_val:
            return max_val
        return value
    except (ValueError, TypeError):
        return default


WEB_URL: str = os.getenv("DORSAL_WEB_URL", "https://dorsalhub.com")
DOCS_URL: str = os.getenv("DORSAL_DOCS_URL", "https://docs.dorsalhub.com")
BASE_URL: str = os.getenv("DORSAL_API_URL", "https://api.dorsalhub.com")

API_MAX_BATCH_SIZE = 1000
ANNOTATION_MAX_SIZE_BYTES = 1024 * 1024
ANNOTATION_SCHEMA_LIMIT_STRICT = 64  # Limit on the number of annotations *per schema* on a `FileRecordStrict`

# == Auth & Config ==
ENV_DORSAL_API_KEY_STR = "DORSAL_API_KEY"
ENV_DORSAL_API_TIMEOUT = "DORSAL_API_TIMEOUT"
ENV_DORSAL_CACHE_ENABLED = "DORSAL_CACHE_ENABLED"
ENV_DORSAL_CACHE_COMPRESSION = "DORSAL_CACHE_COMPRESSION"

# == Timeout Configuration ==
ENV_DORSAL_API_TIMEOUT = "DORSAL_API_TIMEOUT"
_DEFAULT_TIMEOUT = 60.0
API_TIMEOUT = get_float_envvar(ENV_DORSAL_API_TIMEOUT, default=_DEFAULT_TIMEOUT)

# == Batch Size Configuration ==
ENV_DORSAL_BATCH_SIZE = "DORSAL_BATCH_SIZE"
API_MAX_BATCH_SIZE = 1000
API_BATCH_SIZE = get_int_envvar(ENV_DORSAL_BATCH_SIZE, default=API_MAX_BATCH_SIZE, max_val=API_MAX_BATCH_SIZE)

# == DorsalClient Retries ==
ENV_DORSAL_API_MAX_RETRIES = "DORSAL_API_MAX_RETRIES"
_DEFAULT_MAX_RETRIES = 3
API_MAX_RETRIES = get_int_envvar(ENV_DORSAL_API_MAX_RETRIES, default=_DEFAULT_MAX_RETRIES)

LOCAL_DORSAL_DIR = pathlib.Path.home() / ".dorsal"

# Names for the hierarchical config search
PROJECT_CONFIG_FILENAMES = ["dorsal.toml", ".dorsal.toml"]
GLOBAL_CONFIG_FILENAME = "dorsal.toml"
PROJECT_CONFIG_SUBDIR = ".dorsal"

# TOML Sections and Keys
CONFIG_SECTION_AUTH = "auth"
CONFIG_OPTION_API_KEY = "api_key"
CONFIG_OPTION_EMAIL = "email"
CONFIG_OPTION_USER_ID = "user_id"
CONFIG_SECTION_UI = "ui"
CONFIG_OPTION_THEME = "theme"
CONFIG_SECTION_CACHE = "cache"
CONFIG_OPTION_ENABLED = "enabled"
CONFIG_OPTION_COMPRESSION = "compression"


# == CLI ==
CLI_EXPORTS_DIR = LOCAL_DORSAL_DIR / "export"
CLI_SCAN_REPORTS_DIR = LOCAL_DORSAL_DIR / "scan"
CLI_STATS_REPORTS_DIR = LOCAL_DORSAL_DIR / "stats"
CLI_DUPLICATES_REPORTS_DIR = LOCAL_DORSAL_DIR / "duplicates"
CLI_GET_REPORTS_DIR = LOCAL_DORSAL_DIR / "get"
CLI_SEARCH_REPORTS_DIR = LOCAL_DORSAL_DIR / "search"
CLI_SUPPORTED_EXPORT_FORMATS = ["csv", "json"]
CLI_DEFAULT_EXPORT_FORMAT = "json"


# == Docs ==
DOCS_URL_API_TROUBLESHOOTING: str = f"{WEB_URL}/reference/exceptions/"
DOCS_URL_API_ERRORS: str = f"{WEB_URL}/reference/exceptions/"
DOCS_URL_API_AUTH: str = f"{DOCS_URL}/cli/auth/#dorsal-auth-login"
DOCS_URL_API_ERRORS_NETWORK: str = f"{WEB_URL}/reference/exceptions/"
DOCS_URL_API_ERRORS_VALIDATION: str = f"{WEB_URL}/reference/exceptions/"
DOCS_URL_DORSAL_FILE_TAGS: str = f"{WEB_URL}/reference/tags"

# == API ==
API_ENDPOINT_ANNOTATIONS = "v1/annotations"
API_ENDPOINT_COLLECTIONS = "v1/collections"
API_ENDPOINT_EXPORT = "v1/export"
API_ENDPOINT_FILES = "v1/files"
API_ENDPOINT_FILE_SEARCH = "v1/files/search"
API_ENDPOINT_NAMESPACES = "v1/namespaces"
API_ENDPOINT_REGISTRY = "v1/registry"
API_ENDPOINT_FILE_TAG_VALIDATION = "v1/files/tags/validate"
API_ENDPOINT_USER_CHECK_FILES_INDEXED = "v1/users/files-indexed"

# == Core Schemas ==
FILE_BASE_ANNOTATION_SCHEMA = "file/base"
CORE_EBOOK_ANNOTATION_SCHEMA = "file/ebook"
CORE_PDF_ANNOTATION_SCHEMA = "file/pdf"
CORE_MEDIAINFO_ANNOTATION_SCHEMA = "file/mediainfo"
CORE_OFFICE_DOCUMENT_ANNOTATION_SCHEMA = "file/office"
CORE_FILE_DATASETS = {
    FILE_BASE_ANNOTATION_SCHEMA,
    CORE_EBOOK_ANNOTATION_SCHEMA,
    CORE_PDF_ANNOTATION_SCHEMA,
    CORE_MEDIAINFO_ANNOTATION_SCHEMA,
    CORE_OFFICE_DOCUMENT_ANNOTATION_SCHEMA,
}

VALID_DATASET_TYPES = ["File", "Reference"]


# == Open Validation Schemas ===
OPEN_VALIDATION_SCHEMAS_VER = "0.5.0"
ENV_DORSAL_OPEN_VALIDATION_SCHEMAS_DIR = "DORSAL_OPEN_VALIDATION_SCHEMAS_DIR"



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
from typing import Any, Dict

from rich.panel import Panel
from rich.table import Table
from rich.text import Text
from rich.console import Group, RenderableType
from rich.markup import escape
from rich.rule import Rule
import typer


def _get_max_key_width(data: Dict, indent_level: int = 0, display_fields: set | None = None) -> int:
    """Recursively calculates the maximum width needed for keys in a nested dict."""
    max_width = 0
    indent = "  " * indent_level
    for key, value in data.items():
        if key == "file_hash":
            continue
        if display_fields and key not in display_fields:
            continue
        key_width = len(f"{indent}{key}:")
        if key_width > max_width:
            max_width = key_width

        if isinstance(value, dict):
            nested_width = _get_max_key_width(value, indent_level + 1, display_fields=display_fields)
            if nested_width > max_width:
                max_width = nested_width
    return max_width


def _build_annotation_table(
    key: str,
    table: Table,
    data: dict | list[dict],
    level: int = 0,
    is_stub: bool = False,
    display_fields: set | None = None,
):
    """Recursive. Renders annotation stubs in a simplified format."""
    indent = "  " * level

    if isinstance(data, dict):
        if is_stub:
            source: dict = data.get("source", {})
            source_type = source.get("type", "none")
            source_id = source.get("id", "none")

            table.add_row(f"{indent}Source:", f"{source_type} ({source_id})")
            if date_mod := data.get("date_modified"):
                dt_obj = datetime.datetime.fromisoformat(date_mod.replace("Z", "+00:00"))
                table.add_row(f"{indent}Modified:", dt_obj.strftime("%Y-%m-%d %H:%M"))

            annotation_id = data.get("id")
            if annotation_id:
                table.add_row(f"{indent}ID:", annotation_id)

            if data.get("url") and data.get("id"):
                file_hash = data["url"].split("/")[2]
                anno_id = data["id"]
                cta = f"dorsal annotation get {file_hash} {anno_id}"
                table.add_row(f"{indent}To View:", Text(cta, style="cyan"))
        else:
            for key, value in data.items():
                if key == "file_hash":
                    continue

                if display_fields and key not in display_fields:
                    continue

                if not value:
                    continue

                key_text = f"{indent}{key}:"

                if isinstance(value, dict):
                    table.add_row(key_text, "")
                    _build_annotation_table(key, table, value, level + 1, display_fields=display_fields)

                elif isinstance(value, list):
                    if all(not isinstance(item, (dict, list)) for item in value):
                        value_str = ", ".join(str(item) for item in value)
                        table.add_row(key_text, value_str)
                    else:
                        table.add_row(key_text, "")
                        _build_annotation_table(key, table, value, level + 1, display_fields=display_fields)

                else:
                    table.add_row(key_text, str(value))

    elif isinstance(data, list):
        for item in data:
            _build_annotation_table(key, table, item, level, is_stub=is_stub, display_fields=display_fields)


def create_file_info_panel(
    *,
    record_dict: dict,
    title: str,
    private: bool | None,
    palette: dict[str, str],
    override_title_style: str | None = None,
    override_border_style: str | None = None,
    source: str | None = None,
) -> Panel:
    """
    Creates a standardized rich Panel to display file information using a color palette.
    """
    from dorsal.file.utils.size import human_filesize
    from dorsal.cli.views.mediainfo import MEDIAINFO_DISPLAY_FIELDS

    if override_border_style:
        border = override_border_style
    else:
        border = palette["panel_border_alt"] if private else palette["panel_border"]
    if override_title_style:
        panel_title_style = override_title_style
    else:
        panel_title_style = palette["panel_title_alt"] if private else palette["panel_title"]
    primary_color = palette["primary_value_alt"] if private else palette["primary_value"]

    display_title = title
    if source == "cache":
        display_title = f"{title} [{palette.get('info', 'dim')}](from cache)[/]"

    access_text = None
    if private:
        access_text = Text("Private Record", style=palette["access_private"])

    renderables: list[RenderableType] = []

    hashes_table = Table(box=None, show_header=False, padding=(0, 1))
    hashes_table.add_column(style=palette["key"], justify="right", width=12)
    hashes_table.add_column(style=palette["hash_value"])
    if record_dict.get("hash"):
        hashes_table.add_row("SHA-256:", record_dict["hash"])
    if record_dict.get("validation_hash"):
        hashes_table.add_row("BLAKE3:", record_dict["validation_hash"])
    if record_dict.get("quick_hash"):
        hashes_table.add_row("QUICK:", record_dict["quick_hash"])
    if record_dict.get("similarity_hash"):
        hashes_table.add_row("TLSH:", record_dict["similarity_hash"])
    renderables.append(Group(Text.from_markup(f"[{palette['section_title']}]🔑 Hashes[/]"), hashes_table))
    renderables.append(Text(""))

    file_info_table = Table(box=None, show_header=False, padding=(0, 1))
    file_info_table.add_column(style=palette["key"], justify="right", width=12)
    file_info_table.add_column(style=primary_color)
    local_info: dict[str, Any] = record_dict.get("local_filesystem", {})
    base_info = record_dict.get("annotations", {}).get("file/base", {}).get("record", {})
    if access_text:
        file_info_table.add_row("Access:", access_text)
    if local_info.get("is_symlink"):
        link_style = f"{palette['primary_value']} italic"
        target_style = palette["primary_value"]

        file_info_table.add_row("Type:", Text("Symbolic Link", style=link_style))

        if local_info.get("full_path"):
            file_info_table.add_row("Link Path:", escape(local_info["full_path"]))

        target = local_info.get("symlink_target", "Unknown")
        file_info_table.add_row("Target:", Text(target, style=target_style))

    elif local_info.get("full_path"):
        file_info_table.add_row("Full Path:", escape(local_info["full_path"]))
    if local_info.get("date_modified"):
        file_info_table.add_row(
            "Modified:",
            datetime.datetime.fromisoformat(local_info["date_modified"]).strftime("%Y-%m-%d %H:%M:%S"),
        )
    if base_info:
        file_info_table.add_row("Name:", base_info.get("name"))
        file_info_table.add_row("Size:", human_filesize(base_info.get("size", 0)))
        file_info_table.add_row("Media Type:", base_info.get("media_type"))
    renderables.append(
        Group(
            Text.from_markup(f"[{palette['section_title']}]📄 File Info[/]"),
            file_info_table,
        )
    )
    renderables.append(Text(""))

    tags = record_dict.get("tags", [])
    tags_table = Table(box=None, show_header=False, padding=(0, 2))
    tags_table.add_column(style=palette["key"], justify="right")
    tags_table.add_column()
    if tags:
        for tag in tags:
            tag_style = palette["tag_private"] if tag.get("private") else palette["tag_public"]
            tag_value = Text.assemble(
                (f"{tag.get('value')} ", tag_style),
                (f"({'private' if tag.get('private') else 'public'})", palette["tag_subtext"]),
            )
            tag_id = tag.get("id")
            if tag_id:
                tag_value.append(f" [id: {tag_id}]", style=palette["tag_subtext"])
            tags_table.add_row(f"{tag.get('name')}:", tag_value)
    else:
        tags_table.add_row("", Text("No tags found.", style=palette["info"]))
    renderables.append(Group(Text.from_markup(f"[{palette['section_title']}]🔖 Tags[/]"), tags_table))
    renderables.append(Text(""))

    annotations = record_dict.get("annotations", {})
    if annotations:
        for key, value in annotations.items():
            if key == "file/base" or value is None:
                continue

            items_to_render = value if isinstance(value, list) else [value]
            if not items_to_render:
                continue

            first_item = items_to_render[0]
            is_stub = isinstance(first_item, dict) and "record" not in first_item

            title_part = key.split("/")[-1]
            anno_title = f"{title_part.replace('_', ' ').title()} Info"

            renderables.append(
                Text.from_markup(f"[{palette['section_title']}]✏️ {anno_title}[/][{palette['key']}] → {key}[/]")
            )

            if is_stub:
                for i, item in enumerate(items_to_render):
                    if not isinstance(item, dict):
                        continue

                    if i > 0:
                        renderables.append(Rule(style=palette.get("info", "dim")))

                    annotation_table = Table.grid(padding=(0, 1, 0, 2), expand=False)
                    annotation_table.add_column(style=palette["key"], justify="right", width=12)
                    annotation_table.add_column(style=palette["primary_value"])
                    _build_annotation_table(key=key, table=annotation_table, data=item, is_stub=True)
                    renderables.append(annotation_table)
            else:
                record_to_render = items_to_render[0].get("record", items_to_render[0])
                if not isinstance(record_to_render, dict):
                    continue

                display_fields = MEDIAINFO_DISPLAY_FIELDS if key == "file/mediainfo" else None
                max_width = _get_max_key_width(record_to_render, display_fields=display_fields)
                annotation_table = Table(box=None, show_header=False, padding=(0, 1))
                annotation_table.add_column(style=palette["key"], justify="right", min_width=max_width + 2)
                annotation_table.add_column(style=palette["primary_value"])
                _build_annotation_table(
                    key=key, table=annotation_table, data=record_to_render, display_fields=display_fields
                )
                renderables.append(annotation_table)

            renderables.append(Text(""))

    return Panel(
        Group(*renderables),
        title=f"[{panel_title_style}]{display_title}[/]",
        border_style=border,
        expand=False,
        padding=(1, 2),
    )



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import typer
from typing import Annotated

logger = logging.getLogger(__name__)


def clear_cache(
    ctx: typer.Context,
    yes: Annotated[
        bool,
        typer.Option(
            "--yes",
            "-y",
            help="Bypass the confirmation prompt and clear the cache immediately.",
        ),
    ] = False,
):
    """
    Deletes the entire cache database file, permanently removing all cached data.
    """
    from dorsal.common.cli import get_rich_console, exit_cli, EXIT_CODE_ERROR
    from dorsal.session import get_shared_cache, clear_shared_cache

    console = get_rich_console()

    palette = ctx.obj["palette"]

    try:
        cache = get_shared_cache()
        cache_path_str = str(cache.db_path.resolve())

        if not yes:
            typer.confirm(
                f"Are you sure you want to delete the entire cache at {cache_path_str}?\nThis action cannot be undone.",
                abort=True,
            )

        with console.status("Clearing cache..."):
            cache.clear()
            clear_shared_cache()

        success_msg = f"[{palette.get('success', 'green')}]✅ Cache cleared successfully.[/]"
        console.print(success_msg)

    except typer.Abort:
        console.print("Cache clearing cancelled.")
        exit_cli()
    except typer.Exit:
        raise
    except Exception as err:
        logger.exception("Failed to clear cache.")
        exit_cli(
            code=EXIT_CODE_ERROR,
            message=f"An error occurred while clearing the cache: {err}",
        )



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
import gzip
import json
import sqlite3
import os
import logging
import weakref
from pathlib import Path
from typing import Literal, TYPE_CHECKING
import zlib

from pydantic import BaseModel, Field

if TYPE_CHECKING:
    from dorsal.file.validators.file_record import FileRecordStrict

logger = logging.getLogger(__name__)


class CachedFileRecord(BaseModel):
    """Pydantic model representing a single record in the cache database."""

    abspath: str
    modified_time: float
    record_json: str = Field(alias="record")
    name: str | None = None
    size: int | None = None
    extension: str | None = None
    media_type: str | None = None
    hash_sha256: str
    hash_blake3: str | None = None
    hash_quick: str | None = None
    hash_tlsh: str | None = None


class DorsalCache:
    """
    Manages the local SQLite cache for file metadata.
    """

    def __init__(self, db_path: Path | None = None, use_compression: bool = True):
        if db_path:
            self.db_path = db_path
        else:
            self.db_path = Path.home() / ".dorsal" / "cache.db"

        self.use_compression = use_compression
        self.conn: sqlite3.Connection | None = None
        self._ensure_db_directory_exists()
        self._finalizer = weakref.finalize(self, self._finalize_connection, self.conn)
        logger.debug(f"DorsalCache initialized for path: {self.db_path} with compression={self.use_compression}")

    @staticmethod
    def _finalize_connection(conn):
        if conn:
            try:
                conn.close()
            except Exception:
                pass

    def _ensure_db_directory_exists(self):
        """Ensures the parent directory for the database file exists."""
        if not self.db_path.parent.exists():
            logger.debug(f"Creating cache directory at: {self.db_path.parent}")
            self.db_path.parent.mkdir(parents=True, exist_ok=True)

    def connect(self):
        """Establishes a connection to the SQLite database and initializes the schema."""
        if self.conn is None:
            logger.debug(f"Connecting to cache database: {self.db_path}")
            self.conn = sqlite3.connect(self.db_path)
            self.conn.row_factory = sqlite3.Row
            self._finalizer = weakref.finalize(self, self._finalize_connection, self.conn)
            self._initialize_schema()

    def _ensure_connection(self) -> sqlite3.Connection:
        """Ensures the database connection is active, returning the connection object."""
        if self.conn is None:
            self.connect()
        if self.conn is None:
            raise RuntimeError("Database connection could not be established.")
        return self.conn

    def _initialize_schema(self):
        """Creates/updates the `cached_files` table and its indexes."""
        conn = self._ensure_connection()
        cursor = conn.cursor()

        logger.debug("Initializing cache schema...")
        cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS cached_files (
                abspath TEXT PRIMARY KEY,
                modified_time REAL NOT NULL,
                record BLOB,
                is_compressed INTEGER DEFAULT 0,
                name TEXT,
                size INTEGER,
                extension TEXT,
                media_type TEXT,
                hash_sha256 TEXT,
                hash_blake3 TEXT,
                hash_quick TEXT,
                hash_tlsh TEXT
            );
            """
        )

        logger.debug("Ensuring all cache indexes exist...")
        cursor.execute("CREATE INDEX IF NOT EXISTS idx_hash_sha256 ON cached_files (hash_sha256);")
        cursor.execute("CREATE INDEX IF NOT EXISTS idx_hash_blake3 ON cached_files (hash_blake3);")
        cursor.execute("CREATE INDEX IF NOT EXISTS idx_hash_quick ON cached_files (hash_quick);")
        cursor.execute("CREATE INDEX IF NOT EXISTS idx_hash_tlsh ON cached_files (hash_tlsh);")
        cursor.execute("CREATE INDEX IF NOT EXISTS idx_name ON cached_files (name);")
        cursor.execute("CREATE INDEX IF NOT EXISTS idx_extension ON cached_files (extension);")
        cursor.execute("CREATE INDEX IF NOT EXISTS idx_media_type ON cached_files (media_type);")

        conn.commit()
        logger.debug("Schema initialization complete.")

    def get_record(self, *, path: str) -> CachedFileRecord | None:
        """Retrieves a record, decompressing it if necessary."""
        conn = self._ensure_connection()
        logger.debug(f"Attempting to get record for path: {path}")
        cursor = conn.cursor()
        cursor.execute(
            """
            SELECT abspath, modified_time, record, is_compressed, name, size,
                   extension, media_type, hash_sha256, hash_blake3,
                   hash_quick, hash_tlsh
            FROM cached_files WHERE abspath = ?
            """,
            (path,),
        )
        row = cursor.fetchone()

        if not row:
            logger.debug(f"Cache miss for path: {path}")
            return None

        logger.debug(f"Cache hit for path: {path}")
        row_dict = dict(row)
        record_data: bytes | None = row_dict["record"]
        is_compressed_flag = row_dict["is_compressed"]

        if record_data is None:
            logger.debug(f"Cache entry for path '{path}' has NULL data. Treating as a cache miss.")
            return None

        if is_compressed_flag:
            logger.debug(f"Decompressing record for path: {path}")
            record_json_str = zlib.decompress(record_data).decode("utf-8")
        else:
            record_json_str = record_data.decode("utf-8")

        row_dict["record"] = record_json_str
        return CachedFileRecord.model_validate(row_dict)

    def upsert_record(self, *, path: str, modified_time: float, record: "FileRecordStrict"):
        """Inserts or replaces a record, respecting the compression setting."""
        conn = self._ensure_connection()
        logger.debug(f"Upserting record for path: {path}")
        base_annotation = record.annotations.file_base.record
        all_hashes = base_annotation.all_hash_ids or {}

        record_json_str = record.model_dump_json()
        is_compressed_flag = 0

        if self.use_compression:
            logger.debug(f"Compressing record for path: {path}")
            record_data = zlib.compress(record_json_str.encode("utf-8"))
            is_compressed_flag = 1
        else:
            record_data = record_json_str.encode("utf-8")

        sql_data = {
            "abspath": path,
            "modified_time": modified_time,
            "record": record_data,
            "is_compressed": is_compressed_flag,
            "name": base_annotation.name,
            "size": base_annotation.size,
            "extension": base_annotation.extension,
            "media_type": base_annotation.media_type,
            "hash_sha256": all_hashes.get("SHA-256"),
            "hash_blake3": all_hashes.get("BLAKE3"),
            "hash_quick": all_hashes.get("QUICK"),
            "hash_tlsh": all_hashes.get("TLSH"),
        }

        cursor = conn.cursor()
        cursor.execute(
            """
            INSERT OR REPLACE INTO cached_files (
                abspath, modified_time, record, is_compressed, name, size,
                extension, media_type, hash_sha256, hash_blake3,
                hash_quick, hash_tlsh
            ) VALUES (
                :abspath, :modified_time, :record, :is_compressed, :name, :size,
                :extension, :media_type, :hash_sha256, :hash_blake3,
                :hash_quick, :hash_tlsh
            )
            """,
            sql_data,
        )
        conn.commit()
        logger.debug(f"Successfully upserted record for path: {path}")

    def upsert_hash(self, *, path: str, modified_time: float, hash_function: str, hash_value: str):
        """
        Inserts or updates a single hash, correctly handling and invalidating
        existing full records if they are stale.
        """
        conn = self._ensure_connection()
        field_map = {
            "SHA-256": "hash_sha256",
            "BLAKE3": "hash_blake3",
            "QUICK": "hash_quick",
            "TLSH": "hash_tlsh",
        }
        column_name = field_map.get(hash_function.upper())
        if not column_name:
            raise ValueError(f"Unsupported hash function '{hash_function}'.")

        cursor = conn.cursor()
        cursor.execute("SELECT record IS NOT NULL FROM cached_files WHERE abspath = ?", (path,))
        result = cursor.fetchone()
        if result and result[0]:
            stale_check_sql = "SELECT modified_time FROM cached_files WHERE abspath = ?"
            cursor.execute(stale_check_sql, (path,))
            cached_mod_time = cursor.fetchone()[0]
            if cached_mod_time != modified_time:
                logger.debug(f"Stale full record found for '{path}'. Deleting before upserting new hash.")
                cursor.execute("DELETE FROM cached_files WHERE abspath = ?", (path,))

        sql = f"""
            INSERT INTO cached_files (abspath, modified_time, {column_name})
            VALUES (?, ?, ?)
            ON CONFLICT(abspath) DO UPDATE SET
                modified_time = excluded.modified_time,
                {column_name} = excluded.{column_name};
        """
        cursor.execute(sql, (path, modified_time, hash_value))
        conn.commit()

    def get_hash(self, *, path: str, hash_function: str = "SHA-256") -> str | None:
        """
        Efficiently retrieves a specific hash for a cached file if the cache is valid.
        """
        conn = self._ensure_connection()
        field_map = {
            "SHA-256": "hash_sha256",
            "BLAKE3": "hash_blake3",
            "QUICK": "hash_quick",
            "TLSH": "hash_tlsh",
        }
        column_name = field_map.get(hash_function.upper())
        if not column_name:
            raise ValueError(f"Unsupported hash function '{hash_function}'. Supported: {list(field_map.keys())}")

        logger.debug(f"Checking cache for '{hash_function}' hash for path: {path}")
        cursor = conn.cursor()

        sql_query = f"SELECT modified_time, {column_name} FROM cached_files WHERE abspath = ?"
        cursor.execute(sql_query, (path,))
        row = cursor.fetchone()

        if not row:
            logger.debug(f"Cache miss: No record found for path: {path}")
            return None

        cached_mod_time = row["modified_time"]
        try:
            current_mod_time = os.lstat(path).st_mtime
        except FileNotFoundError:
            logger.debug(f"Cache stale: File not found on disk at path: {path}")
            return None

        if cached_mod_time != current_mod_time:
            logger.debug(f"Cache miss: Record is stale for path: {path} (mtime mismatch)")
            return None

        hash_value = row[column_name]
        if hash_value:
            logger.debug(f"Cache hit: Found valid '{hash_function}' hash for path: {path}")
            return hash_value
        else:
            logger.debug(f"Cache miss: Record found but missing '{hash_function}' hash for path: {path}")
            return None

    def clear(self):
        """Close the connection and deletes the entire database file."""
        logger.debug(f"Clearing cache by deleting database file: {self.db_path}")
        self.close()
        try:
            if os.path.exists(self.db_path):
                os.remove(self.db_path)
                logger.debug("Cache database file successfully removed.")
        except OSError as e:
            logger.error(f"Error removing cache file at {self.db_path}: {e}")

    def close(self):
        """Commit any pending changes and close the connection."""
        if self.conn:
            logger.debug("Closing cache database connection.")
            self.conn.commit()
            self.conn.close()
            self.conn = None

    def summary(self) -> dict:
        """Provides a summary of the cache's current state."""
        conn = self._ensure_connection()
        logger.debug("Generating cache summary...")
        cursor = conn.cursor()
        cursor.execute("SELECT COUNT(*) FROM cached_files")
        record_count = cursor.fetchone()[0]

        cursor.execute("SELECT COUNT(*) FROM cached_files WHERE record IS NOT NULL")
        full_records = cursor.fetchone()[0]

        hash_only_records = record_count - full_records

        try:
            db_size_bytes = os.path.getsize(self.db_path)
        except FileNotFoundError:
            db_size_bytes = 0

        summary_data = {
            "database_path": str(self.db_path),
            "total_records": record_count,
            "full_records": full_records,
            "hash_only_records": hash_only_records,
            "database_size_bytes": db_size_bytes,
        }
        logger.debug(f"Cache summary generated: {summary_data}")
        return summary_data

    def prune(self) -> tuple[int, int]:
        """Prunes the cache by removing stale records."""
        conn = self._ensure_connection()
        logger.debug("Starting cache prune operation...")
        cursor = conn.cursor()
        cursor.execute("SELECT abspath, modified_time FROM cached_files")
        records = list(cursor.fetchall())
        total_records = len(records)
        logger.debug(f"Scanning {total_records} records for staleness...")

        stale_paths = []
        for record in records:
            path, cached_mod_time = record["abspath"], record["modified_time"]

            if not os.path.exists(path):
                logger.debug(f"Marking stale (path not found): {path}")
                stale_paths.append(path)
                continue

            try:
                current_mod_time = os.lstat(path).st_mtime
                if current_mod_time != cached_mod_time:
                    logger.debug(f"Marking stale (mtime mismatch): {path}")
                    stale_paths.append(path)
            except FileNotFoundError:
                logger.debug(f"Marking stale (path disappeared during check): {path}")
                stale_paths.append(path)

        if not stale_paths:
            logger.debug("Prune complete. No stale records found.")
            return (0, total_records)

        logger.debug(f"Removing {len(stale_paths)} stale records...")
        cursor.executemany(
            "DELETE FROM cached_files WHERE abspath = ?",
            [(path,) for path in stale_paths],
        )
        conn.commit()

        logger.debug(f"Prune complete. Removed {len(stale_paths)} of {total_records} records.")
        return (len(stale_paths), total_records)

    def vacuum(self) -> None:
        """Rebuilds the database file, reclaiming free space."""
        conn = self._ensure_connection()
        logger.debug("Starting cache vacuum...")
        conn.execute("VACUUM")
        conn.commit()
        logger.debug("Cache vacuum complete.")

    def optimize(self) -> dict:
        """Runs a full maintenance routine on the cache."""
        self._ensure_connection()
        logger.debug("Starting full cache optimization...")
        size_before = os.path.getsize(self.db_path) if os.path.exists(self.db_path) else 0
        pruned_count, _ = self.prune()
        rewritten_count = self._sync_compression()
        self.vacuum()
        size_after = os.path.getsize(self.db_path) if os.path.exists(self.db_path) else 0

        result = {
            "stale_records_removed": pruned_count,
            "records_rewritten_for_compression": rewritten_count,
            "size_before_bytes": size_before,
            "size_after_bytes": size_after,
            "size_reclaimed_bytes": size_before - size_after,
        }
        logger.debug(f"Cache optimization complete: {result}")
        return result

    def _sync_compression(self) -> int:
        """Internal helper to re-compress/de-compress records in a memory-efficient way."""
        conn = self._ensure_connection()
        logger.debug("Starting compression sync...")
        read_cursor = conn.cursor()
        write_cursor = conn.cursor()
        read_cursor.execute("SELECT abspath, record, is_compressed FROM cached_files")

        rewritten_count = 0
        for row in read_cursor:
            path, data, is_compressed = (
                row["abspath"],
                row["record"],
                row["is_compressed"],
            )

            if data is None:
                continue

            new_data = None
            if self.use_compression and not is_compressed:
                logger.debug(f"Compressing record during sync: {path}")
                new_data = zlib.compress(data)
                is_compressed_flag = 1
            elif not self.use_compression and is_compressed:
                logger.debug(f"Decompressing record during sync: {path}")
                new_data = zlib.decompress(data)
                is_compressed_flag = 0

            if new_data is not None:
                write_cursor.execute(
                    "UPDATE cached_files SET record = ?, is_compressed = ? WHERE abspath = ?",
                    (new_data, is_compressed_flag, path),
                )
                rewritten_count += 1

        if rewritten_count > 0:
            conn.commit()
            logger.debug(f"Synced compression state for {rewritten_count} records.")
        else:
            logger.debug("All records already match the current compression setting.")

        return rewritten_count

    def export(
        self,
        output_path: Path,
        format: Literal["json", "json.gz"] = "json.gz",
        include_records: bool = True,
    ) -> int:
        """
        Exports the contents of the cache to a file.

        Args:
            output_path: The path to save the exported file.
            format: The desired output format. Defaults to "json.gz".
            include_records: Whether to include the full metadata records.
                             Defaults to True.

        Returns:
            The total number of records exported.
        """
        logger.debug(f"Starting cache export to '{output_path}' in '{format}' format.")

        try:
            output_path.parent.mkdir(parents=True, exist_ok=True)
        except (IOError, OSError) as err:
            logger.error(f"Output path '{output_path}' is not writable: {err}")
            raise IOError(f"Output path '{output_path}' is not writable.") from err

        conn = self._ensure_connection()
        cursor = conn.cursor()

        columns = [
            "abspath",
            "modified_time",
            "name",
            "size",
            "extension",
            "media_type",
            "hash_sha256",
            "hash_blake3",
            "hash_quick",
            "hash_tlsh",
        ]
        if include_records:
            columns.extend(["record", "is_compressed"])

        query = f"SELECT {', '.join(columns)} FROM cached_files"
        cursor.execute(query)

        rows = cursor.fetchall()
        total_records = len(rows)
        logger.debug(f"Fetched {total_records} records from the cache database.")

        data_to_export = []
        for row in rows:
            row_dict = dict(row)
            if include_records and row_dict.get("record"):
                record_data: bytes = row_dict["record"]
                is_compressed = row_dict["is_compressed"]
                try:
                    if is_compressed:
                        record_json_str = zlib.decompress(record_data).decode("utf-8")
                    else:
                        record_json_str = record_data.decode("utf-8")
                    row_dict["record"] = json.loads(record_json_str)
                except (zlib.error, json.JSONDecodeError, UnicodeDecodeError) as e:
                    logger.warning(f"Could not decode record for {row_dict['abspath']}: {e}")
                    row_dict["record"] = {"error": "Could not decode record"}
            data_to_export.append(row_dict)

        try:
            if format == "json.gz":
                with gzip.open(output_path, "wt", encoding="utf-8") as f:
                    json.dump(data_to_export, f, indent=2, default=str)
            elif format == "json":
                with open(output_path, "w", encoding="utf-8") as f:
                    json.dump(data_to_export, f, indent=2, default=str)
            else:
                raise ValueError(f"Unsupported export format: {format}")

            logger.info(f"Successfully exported {total_records} records to {output_path}")
            return total_records

        except (IOError, ValueError):
            logger.exception(f"Failed to write cache export to '{output_path}'.")
            raise



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import datetime
import pathlib
import typer
import pytest
from typer.testing import CliRunner
from unittest.mock import MagicMock, patch, ANY
from rich.panel import Panel
from rich.table import Table


from dorsal.cli import app
from dorsal.cli.dir_app import scan_dir_cmd

runner = CliRunner()

TEST_DATA_DIR = "tests/data"


@pytest.fixture
def mock_rich_console(mocker):
    """Mocks the rich console to capture output."""
    mock_console = MagicMock()
    mocker.patch("dorsal.common.cli.get_rich_console", return_value=mock_console)
    return mock_console


@pytest.fixture
def mock_exit_cli(mocker):
    """
    Mocks exit_cli to ensure it raises typer.Exit with the correct code.
    This fixes the 'assert 0 != 0' failures by propagating the error code.
    """

    def _side_effect(code=0, message=None):
        raise typer.Exit(code)

    return mocker.patch("dorsal.common.cli.exit_cli", side_effect=_side_effect)


@pytest.fixture
def mock_scan_dir_cmd(mocker):
    """Mocks backend dependencies for the `dir scan` command."""
    # Patch LocalFileCollection at its source
    mock_collection_class = mocker.patch("dorsal.file.collection.local.LocalFileCollection")

    # Configure the instance that will be returned by the constructor
    mock_instance = mock_collection_class.return_value
    mock_instance.warnings = []

    # Setup default behaviors
    mock_instance.__len__.return_value = 2
    mock_instance.__bool__.side_effect = lambda: mock_instance.__len__() > 0

    dt1 = datetime.datetime(2023, 1, 1, 10, 0, 0)
    dt2 = datetime.datetime(2023, 1, 2, 10, 0, 0)

    file_1 = MagicMock()
    file_1.name = "file1.txt"
    file_1.size = 512
    file_1.media_type = "text/plain"
    file_1.date_modified = dt1

    file_2 = MagicMock()
    file_2.name = "file2.txt"
    file_2.size = 1024
    file_2.media_type = "application/json"
    file_2.date_modified = dt2

    mock_instance.info.return_value = {
        "overall": {
            "total_files": 2,
            "total_size": 1536,
            "newest_file": {"path": "file2.txt", "date": dt2},
            "oldest_file": {"path": "file1.txt", "date": dt1},
        },
        "by_type": [{"type": "text/plain", "count": 1}, {"type": "application/json", "count": 1}],
        "by_source": [{"source": "disk", "count": 2}],
    }

    mock_instance.__iter__.return_value = iter([file_1, file_2])
    mock_instance.to_dict.return_value = [{"name": "file1.txt"}, {"name": "file2.txt"}]
    mock_instance.source_info = {"path": TEST_DATA_DIR}

    return {
        "collection_class": mock_collection_class,
        "collection_instance": mock_instance,
    }


def test_scan_dir_success_default(mock_rich_console, mock_scan_dir_cmd, mock_exit_cli):
    """Tests the default `dir scan` command."""
    normalized_path = os.path.normpath(str(TEST_DATA_DIR))
    result = runner.invoke(app, ["dir", "scan", normalized_path])

    assert result.exit_code == 0

    mock_scan_dir_cmd["collection_class"].assert_called_once_with(
        source=normalized_path,
        console=mock_rich_console,
        palette=ANY,
        recursive=False,
        use_cache=False,
        overwrite_cache=False,
        follow_symlinks=True,
        lazy=False,
    )

    # Verify Summary Panel was printed
    print_calls = mock_rich_console.print.call_args_list
    panels = [call.args[0] for call in print_calls if isinstance(call.args[0], Panel)]
    assert any("Directory Scan Summary" in str(p.title) for p in panels)

    # Verify Table was printed
    tables = [call.args[0] for call in print_calls if isinstance(call.args[0], Table)]
    assert any("File Scan Details" in str(t.title) for t in tables)


def test_scan_dir_json_output(mock_rich_console, mock_scan_dir_cmd, mock_exit_cli):
    """Tests the `dir scan --json` command."""
    result = runner.invoke(app, ["dir", "scan", TEST_DATA_DIR, "--json"])

    assert result.exit_code == 0
    mock_rich_console.print.assert_called()
    mock_exit_cli.assert_called()

    # Capture the raw JSON output
    json_output_str = mock_rich_console.print.call_args_list[0].args[0]
    data = json.loads(json_output_str)

    assert data["scan_metadata"]["total_files_found"] == 2
    assert data["results"][0]["name"] == "file1.txt"


def test_scan_dir_output_to_file_csv(mock_rich_console, mock_scan_dir_cmd, mock_exit_cli):
    """Tests saving the report to a file with --output."""
    output_file = "my_report.csv"
    result = runner.invoke(app, ["dir", "scan", TEST_DATA_DIR, "--output", output_file])

    assert result.exit_code == 0

    # Verify to_csv was called on the collection
    mock_scan_dir_cmd["collection_instance"].to_csv.assert_called_once()
    args, _ = mock_scan_dir_cmd["collection_instance"].to_csv.call_args
    assert output_file in args[0]

    # Verify success message
    assert "CSV report saved" in str(mock_rich_console.print.call_args_list[-1].args[0])


@pytest.mark.parametrize("sort_by, sort_order", [("size", "desc"), ("date", "asc")])
def test_scan_dir_sorting_options(sort_by, sort_order, mock_scan_dir_cmd, mock_rich_console):
    """Tests that sorting options are passed correctly and table is rendered."""
    result = runner.invoke(
        app,
        [
            "dir",
            "scan",
            TEST_DATA_DIR,
            "--sort-by",
            sort_by,
            "--sort-order",
            sort_order,
        ],
    )

    assert result.exit_code == 0

    # Check that a table was actually generated
    print_calls = mock_rich_console.print.call_args_list
    tables = [call.args[0] for call in print_calls if isinstance(call.args[0], Table)]
    assert len(tables) > 0
    assert "File Scan Details" in str(tables[0].title)


def test_scan_dir_invalid_sort_option(mock_exit_cli):
    """Tests that an invalid sorting option causes a graceful exit."""
    result = runner.invoke(app, ["dir", "scan", TEST_DATA_DIR, "--sort-by", "invalid_column"])

    # Expect exit code != 0 (default error code is usually 1)
    assert result.exit_code != 0
    mock_exit_cli.assert_called()
    assert "Invalid sorting option" in mock_exit_cli.call_args.kwargs["message"]


def test_scan_dir_cache_flag_conflict(mock_exit_cli):
    """Tests that using both --use-cache and --skip-cache fails."""
    result = runner.invoke(app, ["dir", "scan", TEST_DATA_DIR, "--use-cache", "--skip-cache"])

    assert result.exit_code != 0
    mock_exit_cli.assert_called()
    assert "cannot be used together" in mock_exit_cli.call_args.kwargs["message"]


def test_scan_dir_collection_init_failure(mock_scan_dir_cmd, mock_exit_cli):
    """Tests graceful failure if the LocalFileCollection fails to initialize."""
    mock_scan_dir_cmd["collection_class"].side_effect = Exception("Failed to access directory")

    result = runner.invoke(app, ["dir", "scan", TEST_DATA_DIR])

    assert result.exit_code != 0
    mock_exit_cli.assert_called()
    assert "An error occurred during file discovery" in mock_exit_cli.call_args.kwargs["message"]


# --- Expanded Coverage Tests ---


def test_scan_dir_recursive_flag(mock_rich_console, mock_scan_dir_cmd):
    """Tests that the recursive flag is correctly passed to the collection."""
    result = runner.invoke(app, ["dir", "scan", TEST_DATA_DIR, "--recursive"])

    assert result.exit_code == 0
    mock_scan_dir_cmd["collection_class"].assert_called_once_with(
        source=ANY,
        console=ANY,
        palette=ANY,
        recursive=True,
        use_cache=ANY,
        overwrite_cache=ANY,
        follow_symlinks=ANY,
        lazy=ANY,
    )


def test_scan_dir_limit_option(mock_rich_console, mock_scan_dir_cmd):
    """Tests that the limit flag affects table rendering."""
    # We set a low limit
    result = runner.invoke(app, ["dir", "scan", TEST_DATA_DIR, "--limit", "1"])

    assert result.exit_code == 0

    # Check the "Showing first X of Y files" message
    printed_text = "".join(str(c.args[0]) for c in mock_rich_console.print.call_args_list)
    assert "Showing first 1 of 2 files" in printed_text


def test_scan_dir_smart_output_unknown_extension(mock_rich_console, mock_scan_dir_cmd):
    """Tests warning when --output has an unknown extension."""
    result = runner.invoke(app, ["dir", "scan", TEST_DATA_DIR, "--output", "report.txt"])

    assert result.exit_code == 0

    # Warning should be printed
    printed_text = "".join(str(call.args[0]) for call in mock_rich_console.print.call_args_list)
    assert "unknown extension" in printed_text

    # Neither save method should be called
    mock_scan_dir_cmd["collection_instance"].to_json.assert_not_called()
    mock_scan_dir_cmd["collection_instance"].to_csv.assert_not_called()


def test_scan_dir_explicit_save_flags(mock_scan_dir_cmd):
    """Tests that --save and --csv flags force saving."""
    # Case 1: -s flag (save JSON)
    runner.invoke(app, ["dir", "scan", TEST_DATA_DIR, "-s"])
    mock_scan_dir_cmd["collection_instance"].to_json.assert_called_once()

    # Case 2: -c flag (save CSV)
    runner.invoke(app, ["dir", "scan", TEST_DATA_DIR, "--csv"])
    mock_scan_dir_cmd["collection_instance"].to_csv.assert_called_once()


def test_scan_dir_empty_directory(mock_rich_console, mock_scan_dir_cmd, mock_exit_cli):
    """Tests behavior when the directory contains no files."""
    # Set length to 0
    mock_scan_dir_cmd["collection_instance"].__len__.return_value = 0

    result = runner.invoke(app, ["dir", "scan", TEST_DATA_DIR])

    assert result.exit_code == 0  # It exits cleanly via exit_cli

    # Verify exit_cli was called (which signifies the command stopped early)
    mock_exit_cli.assert_called()

    # Verify the "processed 0 files" message
    processed_msg = str(mock_rich_console.print.call_args_list[0].args[0])
    assert "processed" in processed_msg

    # Ensure NO Panel or Table was printed (meaning code stopped before _print_directory_summary_panel)
    print_calls = mock_rich_console.print.call_args_list
    panels = [call.args[0] for call in print_calls if isinstance(call.args[0], Panel)]
    tables = [call.args[0] for call in print_calls if isinstance(call.args[0], Table)]

    assert len(panels) == 0, "Summary panel was printed despite empty collection"
    assert len(tables) == 0, "Table was printed despite empty collection"


def test_scan_dir_with_warnings(mock_rich_console, mock_scan_dir_cmd):
    """Tests that the warnings panel is displayed if warnings exist."""
    # Inject warnings
    mock_scan_dir_cmd["collection_instance"].warnings = ["Permission denied: /root"]

    result = runner.invoke(app, ["dir", "scan", TEST_DATA_DIR])

    assert result.exit_code == 0

    # Find the warning panel
    print_calls = mock_rich_console.print.call_args_list
    panels = [call.args[0] for call in print_calls if isinstance(call.args[0], Panel)]
    warning_panel = next((p for p in panels if "Warnings" in str(p.title)), None)

    assert warning_panel is not None
    assert "Permission denied" in str(warning_panel.renderable)


def test_scan_save_errors(mock_rich_console, mock_scan_dir_cmd, mock_exit_cli):
    """Tests error handling during file save operations."""
    # Simulate IOError during JSON save
    mock_scan_dir_cmd["collection_instance"].to_json.side_effect = IOError("Disk full")

    runner.invoke(app, ["dir", "scan", TEST_DATA_DIR, "-s"])
    mock_exit_cli.assert_called()
    assert "Error writing to file" in mock_exit_cli.call_args.kwargs["message"]

    # Reset for next case
    mock_exit_cli.reset_mock()

    # Simulate Generic Exception during CSV save
    mock_scan_dir_cmd["collection_instance"].to_csv.side_effect = Exception("Random Error")

    runner.invoke(app, ["dir", "scan", TEST_DATA_DIR, "--csv"])
    # Should NOT exit cli, but print warning
    assert mock_exit_cli.call_count == 0
    printed = "".join(str(c.args[0]) for c in mock_rich_console.print.call_args_list)
    assert "Could not save CSV report" in printed



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
from typing import Any, Literal
import requests
from pydantic import BaseModel, Field, NonNegativeInt, computed_field

from dorsal.file.validators.file_record import AnnotationGroup, FileRecordDateTime
from dorsal.file.validators.collection import (
    FileCollection,
    SingleCollectionResponse,
    HydratedSingleCollectionResponse,
)
from dorsal.common.validators import Pagination
from dorsal.file.validators.common import SHA256Hash


class TagResult(BaseModel):
    name: str
    namespace: str
    value: Any
    value_code: Any | None = None
    private: bool
    status: str
    detail: str | None = None


class IndexResultAnnotation(BaseModel):
    name: str
    status: str
    detail: str | None = None


class IndexResult(BaseModel):
    file_path: str | None = None
    name: str | None = None
    hash: str
    url: str
    annotations: list[IndexResultAnnotation]
    tags: list[TagResult] | None = None


class FileIndexResponse(BaseModel):
    total: int
    success: int
    error: int
    unauthorized: int
    tag_status: str | None = None
    results: list[IndexResult]
    response: requests.Response | None = Field(default=None, exclude=True)

    model_config = {"arbitrary_types_allowed": True}

    def __repr__(self) -> str:
        """Summary of the index operation."""
        status_parts = []
        if self.success > 0:
            status_parts.append(f"Success: {self.success}")
        if self.error > 0:
            status_parts.append(f"Errors: {self.error}")  # Highlight errors
        if self.unauthorized > 0:
            status_parts.append(f"Unauthorized: {self.unauthorized}")

        summary = ", ".join(status_parts)

        # If there are errors, we list them explicitly
        error_details = ""
        if self.error > 0:
            error_details = "\n  [!] ERRORS FOUND:"
            for res in self.results:
                # check for annotation errors
                failed_anns = [ann for ann in res.annotations if ann.status == "error"]
                if failed_anns:
                    error_details += f"\n      - File: {res.name or res.hash[:8]}"
                    for ann in failed_anns:
                        error_details += f"\n        x Annotation '{ann.name}': {ann.detail}"

        return f"<FileIndexResponse [{summary}]{error_details}>"


class NewDatasetResponse(BaseModel):
    created: str
    version: str
    url: str


class FileTagResponse(BaseModel):
    hash: str
    success: bool
    detail: str | None = None
    tags: list[TagResult] | None = None


class RecordIndexResultItem(BaseModel):
    key: str
    status: str
    url: str | None
    detail: str | None


class RecordIndexResult(BaseModel):
    total: int
    success: int
    error: int
    warnings: list[str] | None
    dataset_id: str
    results: list[RecordIndexResultItem]


class FileDeleteResponse(BaseModel):
    file_exists: NonNegativeInt | None = None
    file_deleted: NonNegativeInt = 0
    file_modified: NonNegativeInt = 0
    tags_exist: NonNegativeInt | None = None
    tags_deleted: NonNegativeInt = 0
    private_tags_deleted: int = 0
    annotations_exist: NonNegativeInt | None = None
    annotations_modified: NonNegativeInt = 0
    annotations_deleted: NonNegativeInt = 0
    private_annotations_deleted: int = 0
    user_file_exist: NonNegativeInt | None = None
    user_files_deleted: NonNegativeInt = 0


class CollectionCreateRequest(BaseModel):
    """For creating a new file collection."""

    name: str = Field(..., min_length=1, max_length=256)
    description: str | None = Field(None, max_length=1024)
    is_private: bool = True
    source: dict[str, Any]


class AddFilesRequest(BaseModel):
    """For adding files to a collection."""

    hashes: list[SHA256Hash] = Field(
        description="A list of file SHA-256 hashes to add to the collection.",
        max_length=10_000,
    )


class AddFilesResponse(BaseModel):
    added_count: int
    duplicate_count: int
    invalid_count: int


class CollectionsResponse(BaseModel):
    """For a paginated list of file collections."""

    records: list[FileCollection]
    pagination: Pagination


class CollectionUpdateRequest(BaseModel):
    """For updating a collection's properties."""

    name: str | None = Field(None, min_length=1, max_length=256)
    description: str | None = Field(None, max_length=1024)


class CollectionsDeleteRequest(BaseModel):
    """For deleting one or more collections."""

    collection_ids: list[str] = Field(..., min_length=1)


class CollectionWebLocationResponse(BaseModel):
    location_url: str


class ExportJobStatus(BaseModel):
    job_id: str
    status: Literal["PENDING", "RUNNING", "COMPLETED", "FAILED", "UNKNOWN"]
    message: str | None = None
    progress: float = Field(default=0.0)
    download_url: str | None = None
    expires_at: datetime.datetime | None = None


class ExportJobRequest(BaseModel):
    format: Literal["json.gz"] = "json.gz"


class RemoveFilesRequest(BaseModel):
    hashes: list[SHA256Hash] = Field(
        description="A list of file SHA-256 hashes to remove from the collection.",
        max_length=10_000,
    )


class RemoveFilesResponse(BaseModel):
    removed_count: int = Field(description="The number of files successfully removed from the collection.")
    not_found_count: int = Field(description="The number of provided hashes that were not found in the collection.")


class CollectionSyncRequest(BaseModel):
    hashes: list[SHA256Hash] = Field(
        description="The complete list of file SHA-256 hashes the collection should contain.",
        max_length=1_000_000,
    )


class CollectionSyncResponse(BaseModel):
    added_count: int = Field(description="The number of new files added to the collection.")
    removed_count: int = Field(description="The number of files removed from the collection.")
    unchanged_count: int = Field(
        description="The number of files that were already in the collection and were not changed."
    )


class CollectionSyncJob(BaseModel):
    job_id: str = Field(description="The unique ID for the synchronization job.")
    status: str = Field(description="The initial status of the job (e.g., PENDING).")


class CollectionSyncJobStatus(CollectionSyncJob):
    result: CollectionSyncResponse | None = Field(
        default=None,
        description="The final result of the sync, available when the job is complete.",
    )
    error: str | None = Field(default=None, description="An error message if the job failed.")


class AnnotationIndexResultItem(BaseModel):
    key: str
    annotation_id: str | None
    private: bool
    status: Literal["created", "updated", "skipped", "error"]
    detail: str | None = None


class AnnotationIndexResult(BaseModel):
    """For adding or updating file annotations."""

    total: int
    success: int
    error: int
    dataset_id: str
    results: list[AnnotationIndexResultItem]


class FileAnnotationResponse(BaseModel):
    annotation_id: str
    file_hash: str
    schema_id: str
    schema_version: str | None = None
    source: dict[str, Any]
    record: dict[str, Any]
    user_id: int
    date_created: datetime.datetime
    date_modified: datetime.datetime
    private: bool


class FileAnnotationGroupResponse(BaseModel):
    annotation_id: str
    file_hash: str
    schema_id: str
    schema_version: str | None = None
    source: dict[str, Any]
    group: AnnotationGroup
    user_id: int
    date_created: datetime.datetime
    date_modified: datetime.datetime
    private: bool


class RegistryModelResponse(BaseModel):
    namespace: str
    name: str
    version: str
    install_url: str
    schema_id: str
    schema_version: str | None = None
    description: str | None = None
    package_name: str
    is_official: bool = False
    is_verified: bool = False
    created_at: datetime.datetime | None = None
    dependencies: list[dict[str, Any]] | None = None
